# -*- coding: utf-8 -*-
"""Python GR module

Exported Classes:

"""

# standard library
import math
import time
import logging
# local library
import gr
import gr3
from gr.pygr.base import GRDrawAttributes, GRMeta, GRViewPort
from gr.pygr.helper import ColorIndexGenerator, DomainChecker
from numpy import ndarray
from gr._version import __version__, __revision__

__author__ = """Christian Felder <c.felder@fz-juelich.de>,
Josef Heinen <j.heinen@fz-juelich.de>"""
__copyright__ = """Copyright 2012-2014 Forschungszentrum Juelich GmbH

This file is part of GR, a universal framework for visualization applications.
Visit http://gr-framework.org for the latest
version.

GR was developed by the Scientific IT-Systems group at the Peter Grünberg
Institute at Forschunsgzentrum Jülich. The main development has been done
by Josef Heinen who currently maintains the software.

GR is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This framework is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GR. If not, see <http://www.gnu.org/licenses/>.
 
"""

_log = logging.getLogger(__name__)

class Point(object):

    def __init__(self, x, y):
        self._x, self._y = x, y

    @property
    def x(self):
        """Get the current x value."""
        return self._x

    @x.setter
    def x(self, value):
        self._x = value

    @property
    def y(self):
        """Get the current y value."""
        return self._y

    @y.setter
    def y(self, value):
        self._y = value

    def __str__(self):
        return "(%s, %s)" % (self._x, self._y)

    def __eq__(self, other):
        return (self.x == other.x and self.y == other.y)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __add__(self, other):
        return Point(self.x + other.x, self.y + other.y)

    def __sub__(self, other):
        return Point(self.x - other.x, self.y - other.y)

    def __mul__(self, other):
        """Calculate scalar product."""
        return self.x * other.x + self.y * other.y

    def __div__(self, other):
        """Calculate component-by-component division."""
        return Point(self.x / other.x, self.y / other.y)

    def __neg__(self):
        """Calculate negation."""
        return Point(-self.x, -self.y)

    def __pos__(self):
        return self

    def __abs__(self):
        return Point(abs(self.x), abs(self.y))

    def norm(self):
        """Calculate euclidean norm."""
        return math.sqrt(self * self)

class RegionOfInterest(object):

    LEGEND = 0x1

    def __init__(self, *args, **kwargs):
        self._poly, self._x, self._y = None, None, None
        self._ref = None
        self._regionType = None
        for p in args:
            self.append(p)
        if "reference" in kwargs:
            self._ref = kwargs["reference"]
        if "regionType" in kwargs:
            self._regionType = kwargs["regionType"]

    @property
    def x(self):
        """Get the current x values."""
        return self._x

    @property
    def y(self):
        """Get the current y values."""
        return self._y

    @property
    def reference(self):
        """Get the current reference."""
        return self._ref

    @reference.setter
    def reference(self, ref):
        self._ref = ref

    @property
    def regionType(self):
        """Get the current regionType."""
        return self._regionType

    @regionType.setter
    def regionType(self, rtype):
        self._regionType = rtype

    def append(self, *args):
        if self._poly is None:
            self._poly = []
            self._x = []
            self._y = []
        for p0 in args:
            self._poly.append(p0)
            self._x.append(p0.x)
            self._y.append(p0.y)

    def isPointInside(self, p0):
        """
        
        """
        # Even-odd rule (Point in Polygon Test)
        # assume a ray from testpoint to infinity along positive x axis
        # count intersections along this ray.
        # If odd inside - outside otherwise.
        res = False
        if self._poly:
            n = len(self._poly)
            j = n - 1
            for i in range(n):
                pi, pj = self._poly[i], self._poly[j]
                # if testpoint y component between pi, pj
                # => intersection possible
                if (((pi.y > p0.y) != (pj.y > p0.y)) and
                     # if x components are both greater than testpoint's x
                     # => intersection (with polyline)
                     (((pi.x > p0.x) and (pj.x > p0.x)) or
                    # if testpoint left from intersection (with polyline)
                    # along x axis with same y component
                    # => intersection with polyline
                       (p0.x < ((p0.y - pi.y) * (pi.x - pj.x) / (pi.y - pj.y) + pi.x))
                     )
                    ):
                    # flip on intersection
                    res = not res
                j = i
        return res

class CoordConverter(object):

    def __init__(self, sizex, sizey, window=None):
        self._sizex, self._sizey, self._window = sizex, sizey, window
        self._p = None # always stored in NDC

    def _checkRaiseXY(self):
        if self._p.x is None or self._p.y is None:
            raise AttributeError("x or y has not been initialized.")
        return True

    @property
    def window(self):
        return self._window

    def setWindow(self, xmin, xmax, ymin, ymax):
        self._window = [xmin, xmax, ymin, ymax]

    def getWindow(self):
        if self._window:
            return list(self._window)
        else:
            return None

    def setNDC(self, x, y):
        self._p = Point(x, y)
        return self

    def setWC(self, x, y, viewport, window=None):
        scale = gr.inqscale()
        if window:
            self.setWindow(*window)
        if self.getWindow():
            xmin, xmax, ymin, ymax = self.getWindow()
        else:
            xmin, xmax, ymin, ymax = gr.inqwindow()
        vp = viewport
        if scale & gr.OPTION_Y_LOG:
            ymin, ymax = math.log10(ymin), math.log10(ymax)
            y = math.log10(y)
        if scale & gr.OPTION_X_LOG:
            xmin, xmax = math.log10(xmin), math.log10(xmax)
            x = math.log10(x)
        ndcX = (vp[0] * self._sizex + (x - xmin) / (xmax - xmin)
                * (vp[1] - vp[0]) * self._sizex)
        ndcY = (vp[2] * self._sizey + (y - ymin) / (ymax - ymin)
                * (vp[3] - vp[2]) * self._sizey)
        self.setNDC(ndcX, ndcY)
        return self

    def setWCforPlotAxes(self, x, y, axes):
        self.setWC(x, y, axes.viewport, axes.getWindow())
        return self

    def getNDC(self):
        self._checkRaiseXY()
        return self._p

    def getWC(self, viewport):
        scale = gr.inqscale()
        self._checkRaiseXY()
        ndcPoint = self.getNDC()
        ndcPoint = Point(ndcPoint.x, ndcPoint.y)
        if self.getWindow():
            xmin, xmax, ymin, ymax = self.getWindow()
        else:
            xmin, xmax, ymin, ymax = gr.inqwindow()
        vp = viewport
        if scale & gr.OPTION_Y_LOG:
            ymin, ymax = math.log10(ymin), math.log10(ymax)
        if scale & gr.OPTION_X_LOG:
            xmin, xmax = math.log10(xmin), math.log10(xmax)
        wcY = (ymin + (ndcPoint.y / self._sizey - vp[2])
               * (ymax - ymin) / (vp[3] - vp[2]))
        wcX = (xmin + (ndcPoint.x / self._sizex - vp[0])
               * (xmax - xmin) / (vp[1] - vp[0]))
        if scale & gr.OPTION_Y_LOG:
            wcY = 10 ** wcY
        if scale & gr.OPTION_X_LOG:
            wcX = 10 ** wcX
        return Point(wcX, wcY)


class DeviceCoordConverter(CoordConverter):

    def __init__(self, width, height, sizex=None, sizey=None, window=None):
        self._width, self._height = width, height
        if sizex is None and sizey is None:
            if self._width > self._height:
                sizex = 1.
                sizey = float(self._height) / self._width
            else:
                sizex = float(self._width) / self._height
                sizey = 1.
        CoordConverter.__init__(self, sizex, sizey, window)

    def setDC(self, x, y):
        self._p = Point(float(x) / self._width * self._sizex,
                        (1. - float(y) / self._height) * self._sizey)
        return self

    def getDC(self):
        self._checkRaiseXY()
        return Point(self._p.x / self._sizex * self._width,
                     (1. - self._p.y / self._sizey) * self._height)


class Text(GRMeta):

    DEFAULT_CHARHEIGHT_FACTOR = .027

    def __init__(self, x, y, text, axes=None,
                 charheight=DEFAULT_CHARHEIGHT_FACTOR, charup=(0, 1),
                 halign=gr.TEXT_HALIGN_LEFT, valign=gr.TEXT_VALIGN_TOP,
                 textcolor=1, font=None, precision=None,
                 textpath=gr.TEXT_PATH_RIGHT, charexpan=1.,
                 hideviewport=True):
        self._font, self._precision = None, None
        self._visible = True
        self._x, self._y, self._text, self._axes = x, y, text, axes
        self._charheight = charheight
        self._charup, self._halign, self._valign = charup, halign, valign
        self._textcolor = textcolor
        self._textpath, self._charexpan = textpath, charexpan
        self._hideviewport = hideviewport
        if font is not None and precision is not None:
            self._font, self._precision = font, precision
        self._textlines = text.split('\n')

    @property
    def x(self):
        """Get the current x value (NDC)."""
        return self._x

    @x.setter
    def x(self, value):
        self._x = value

    @property
    def y(self):
        """Get the current y value (NDC)."""
        return self._y

    @y.setter
    def y(self, value):
        self._y = value

    @property
    def text(self):
        """Return the current text string."""
        return self._text

    @text.setter
    def text(self, text):
        self._text = text
        self._textlines = text.split('\n')

    @property
    def charheight(self):
        """Return the current charheight factor."""
        return self._charheight

    @charheight.setter
    def charheight(self, height):
        self._charheight = height

    @property
    def charup(self):
        """Return the current charup vector (ux, uy)."""
        return self._charup

    @charup.setter
    def charup(self, vector):
        self._charup = vector

    @property
    def halign(self):
        """Return the horizontal alignment."""
        return self._halign

    @halign.setter
    def halign(self, horizontal):
        self._halign = horizontal

    @property
    def valign(self):
        """Return the vertical alignment."""
        return self._valign

    @valign.setter
    def valign(self, vertical):
        self._valign = vertical

    @property
    def textcolor(self):
        """Return the color index for the text color."""
        return self._textcolor

    @textcolor.setter
    def textcolor(self, colorind):
        self._textcolor = colorind

    @property
    def textpath(self):
        """Return the current text path (direction in which text will be drawn)."""
        return self._textpath

    @textpath.setter
    def textpath(self, direction):
        self._textpath = direction

    @property
    def charexpan(self):
        """Return the current character expansion factor (width to height ratio)."""
        return self._charexpan

    @charexpan.setter
    def charexpan(self, factor):
        self._charexpan = factor

    def setFont(self, font, precision=gr.TEXT_PRECISION_STRING):
        """Set the font and the precision of the text representation."""
        self._font, self._precision = font, precision

    @property
    def hideviewport(self):
        """Return the current hide outside viewport state."""
        return self._hideviewport

    @hideviewport.setter
    def hideviewport(self, flag):
        self._hideviewport = flag

    @property
    def visible(self):
        return self._visible

    @visible.setter
    def visible(self, bool):
        self._visible = bool

    def getBoundingBox(self):
        if self._axes:
            coord = CoordConverter(self._axes.sizex, self._axes.sizey,
                                   self._axes.getWindow())
            coord.setWCforPlotAxes(self.x, self.y, self._axes)
            p0 = coord.getNDC()
            x, y = p0.x, p0.y

            charHeight = self.charheight * self._axes.sizey
            window = gr.inqwindow()
            # set viewport and window to NDC to allow 'line-drawing'
            # in all regions and in NDC coordinates
            # if hideviewport is False.
            gr.setviewport(0, self._axes.sizex, 0, self._axes.sizey)
            gr.setwindow(0, self._axes.sizex, 0, self._axes.sizey)
        else:
            x, y = self.x, self.y
            charHeight = self.charheight

        gr.setcharheight(charHeight)
        gr.setcharup(*self.charup)
        gr.settextalign(self.halign, self.valign)
        gr.settextpath(self.textpath)
        gr.setcharexpan(self.charexpan)
        if self._font is not None and self._precision is not None:
            gr.settextfontprec(self._font, self._precision)

        tbx, tby = gr.inqtext(x, y, self.text)

        if self._axes:
            gr.setviewport(*self._axes.viewportscaled)
            gr.setwindow(*window)
        return tbx, tby

    def drawGR(self):
        if self.visible:
            isInViewport = True
            if self._axes:
                coord = CoordConverter(self._axes.sizex, self._axes.sizey,
                                       self._axes.getWindow())
                coord.setWCforPlotAxes(self.x, self.y, self._axes)
                p0 = coord.getNDC()
                x, y = p0.x, p0.y
                charHeight = self.charheight * self._axes.sizey
                window = gr.inqwindow()
                # set viewport and window to NDC to allow 'line-drawing'
                # in all regions and in NDC coordinates
                # if hideviewport is False.
                gr.setviewport(0, self._axes.sizex, 0, self._axes.sizey)
                gr.setwindow(0, self._axes.sizex, 0, self._axes.sizey)
            else:
                x, y = self.x, self.y
                charHeight = self.charheight

            gr.setcharheight(charHeight)
            gr.setcharup(*self.charup)
            gr.settextalign(self.halign, self.valign)
            gr.settextpath(self.textpath)
            gr.setcharexpan(self.charexpan)
            if self._font is not None and self._precision is not None:
                gr.settextfontprec(self._font, self._precision)
            gr.settextcolorind(self.textcolor)

            if self._axes:
                tbx, tby = gr.inqtext(x, y, self.text)
                tbxmin, tbxmax = min(tbx), max(tbx)
                tbymin, tbymax = min(tby), max(tby)
                xmin, xmax, ymin, ymax = self._axes.viewportscaled
                if (tbxmin < xmin or tbxmax > xmax
                    or tbymin < ymin or tbymax > ymax):
                    isInViewport = False

            if not self.hideviewport or isInViewport:
                gr.text(x, y, self.text)

            if self._axes:
                gr.setviewport(*self._axes.viewportscaled)
                gr.setwindow(*window)
            gr.settextcolorind(1)


class ErrorBar(GRDrawAttributes, GRMeta):

    HORIZONTAL = 0
    VERTICAL = 1

    def __init__(self, x, y, dneg, dpos, direction=VERTICAL,
                 linetype=gr.LINETYPE_SOLID, markertype=gr.MARKERTYPE_OMARK,
                 linecolor=1, markercolor=1):
        super(ErrorBar, self).__init__(linetype, markertype, linecolor,
                                       markercolor)
        self._x, self._y, self._dneg, self._dpos = x, y, dneg, dpos
        self._direction = direction
        self._grerror = None
        if direction == ErrorBar.VERTICAL:
            self._grerror = gr.verrorbars
        elif direction == ErrorBar.HORIZONTAL:
            self._grerror = gr.herrorbars
        else:
            raise AttributeError("unsupported value for direction.")

    @property
    def direction(self):
        """Get ErrorBars direction."""
        return self._direction

    @property
    def x(self):
        """Get the current list/ndarray of x values."""
        return self._x

    @x.setter
    def x(self, lst):
        if self.direction == ErrorBar.HORIZONTAL:
            self._dneg = self._dneg - self.x + lst
            self._dpos = self._dpos - self.x + lst
        self._x = lst

    @property
    def y(self):
        """Get the current list/ndarray of y values."""
        return self._y

    @y.setter
    def y(self, lst):
        if self.direction == ErrorBar.VERTICAL:
            self._dneg = self._dneg - self.y + lst
            self._dpos = self._dpos - self.y + lst
        self._y = lst

    def drawGR(self):
        # preserve old values
        ltype = gr.inqlinetype()
        mtype = gr.inqmarkertype()
        lcolor = gr.inqlinecolorind()
        mcolor = gr.inqmarkercolorind()

        if self.linetype is not None:
            gr.setlinecolorind(self.linecolor)
            gr.setmarkercolorind(self.markercolor)
            gr.setlinetype(self.linetype)
        if self.markertype is not None:
            gr.setmarkertype(self.markertype)
        else:
            gr.setmarkertype(gr.MARKERTYPE_DOT)

        self._grerror(self._x, self._y, self._dneg, self._dpos)

        # restore old values
        gr.setlinecolorind(lcolor)
        gr.setmarkercolorind(mcolor)
        gr.setlinetype(ltype)
        gr.setmarkertype(mtype)

class Plot(GRViewPort, GRMeta):

    def __init__(self, viewport=GRViewPort.DEFAULT_VIEWPORT):
        super(Plot, self).__init__(viewport)
        self._lstAxes = []
        self._title, self._subTitle = None, None
        self._lblX, self._lblY = None, None
        self._offsetLblX, self._offsetLblY = 0., 0.
        self._legend, self._legendROI = False, None
        self._autoscale = False
        self._countAxes = 0

    def getAxes(self, idx=None):
        ret = None
        if idx is None:
            ret = self._lstAxes
        elif idx < len(self._lstAxes):
            ret = self._lstAxes[idx]
        return ret

    @property
    def offsetXLabel(self):
        """get NDC Y offset for xlabel"""
        return self._offsetLblX

    @offsetXLabel.setter
    def offsetXLabel(self, yoff):
        self._offsetLblX = yoff

    @property
    def offsetYLabel(self):
        """get NDC X offset for ylabel"""
        return self._offsetLblY

    @offsetYLabel.setter
    def offsetYLabel(self, xoff):
        self._offsetLblY = xoff

    @property
    def xlabel(self):
        """get label for x axis"""
        return self._lblX

    @xlabel.setter
    def xlabel(self, xlabel):
        self._lblX = xlabel

    @property
    def ylabel(self):
        """get label for y axis"""
        return self._lblY

    @ylabel.setter
    def ylabel(self, ylabel):
        self._lblY = ylabel

    @property
    def subTitle(self):
        """get plot subtitle"""
        return self._subTitle

    @subTitle.setter
    def subTitle(self, subTitle):
        self._subTitle = subTitle

    @property
    def title(self):
        """get plot title"""
        return self._title

    @title.setter
    def title(self, title):
        self._title = title

    @GRViewPort.viewport.setter
    def viewport(self, viewport):
        self._viewport = list(viewport)
#        GRViewPort.viewport = viewport
        for axes in self._lstAxes:
            axes.viewport = viewport

    @GRViewPort.sizex.setter
    def sizex(self, value):
        self._sizex = value
        for axes in self._lstAxes:
            axes.sizex = value

    @GRViewPort.sizey.setter
    def sizey(self, value):
        self._sizey = value
        for axes in self._lstAxes:
            axes.sizey = value

    def setLogX(self, bool, rescale=False):
        if bool:
            for axes in self._lstAxes:
                if axes.isXLogDomain():
                    axes.setLogX(bool)
                else:
                    if rescale:
                        lstPlotCurves = axes.getCurves()
                        if lstPlotCurves:
                            visibleCurves = filter(lambda curve: curve.visible,
                                                   lstPlotCurves)
                            if not visibleCurves:
                                visibleCurves = lstPlotCurves
                            window = axes.getWindow()
                            xmax = window[1]
                            xmin = min(map(lambda curve: min(curve.x),
                                           visibleCurves))
                            if xmin <= 0:
                                # 3 constant
                                xmin = 1 / 10 ** max(0, 3 - math.log10(xmax))

                            window[0] = xmin
                            axes.setWindow(*window)
                            axes.setLogX(bool)
                    else:
                        win = axes.getWindow()
                        if win is not None:
                            raise Exception("AXES[%d]: (%d..%d) "
                                            % (axes.getId(), win[0], win[1])
                                            + "not in log(x) domain.")
                        else:
                            raise Exception("AXES[%d] not in log(x) domain."
                                            % axes.getId())
        else:
            for axes in self._lstAxes:
                axes.setLogX(bool)

    def setLogY(self, bool, rescale=False):
        if bool:
            for axes in self._lstAxes:
                if axes.isYLogDomain():
                    axes.setLogY(bool)
                else:
                    if rescale:
                        lstPlotCurves = axes.getCurves()
                        if lstPlotCurves:
                            visibleCurves = filter(lambda curve: curve.visible,
                                                   lstPlotCurves)
                            if not visibleCurves:
                                visibleCurves = lstPlotCurves
                            window = axes.getWindow()
                            ymax = window[3]
                            ymin = min(map(lambda curve: min(curve.y),
                                           visibleCurves))
                            if ymin <= 0:
                                # 3 constant
                                ymin = 1 / 10 ** max(0, 3 - math.log10(ymax))

                            window[2] = ymin
                            axes.setWindow(*window)
                            axes.setLogY(bool)
                    else:
                        win = axes.getWindow()
                        if win is not None:
                            raise Exception("AXES[%d]: (%d..%d) "
                                            % (axes.getId(), win[2], win[3])
                                            + "not in log(y) domain.")
                        else:
                            raise Exception("AXES[%d] not in log(y) domain."
                                            % axes.getId())
        else:
            for axes in self._lstAxes:
                axes.setLogY(bool)

    def setGrid(self, bool):
        for axes in self._lstAxes:
            axes.setGrid(bool)

    def isLegendEnabled(self):
        return self._legend

    def setLegend(self, bool):
        self._legend = bool

    @property
    def autoscale(self):
        return self._autoscale

    @autoscale.setter
    def autoscale(self, bool):
        self._autoscale = bool
        for axes in self._lstAxes:
            axes.autoscale = bool

    def reset(self):
        for axes in self._lstAxes:
            axes.reset()

    def logXinDomain(self):
        logXinDomain = True
        for axes in self._lstAxes:
            logXinDomain = (logXinDomain & axes.isXLogDomain())
        return logXinDomain

    def logYinDomain(self):
        logYinDomain = True
        for axes in self._lstAxes:
            logYinDomain = (logYinDomain & axes.isYLogDomain())
        return logYinDomain

    def pick(self, p0, width, height):
        coord, axes, curve = None, None, None
        window = gr.inqwindow()
        if self._lstAxes:
            coord = DeviceCoordConverter(width, height, self._sizex,
                                         self._sizey)
            points = []
            lstAxes = []
            lstCurves = []
            for axes in self._lstAxes:
                gr.setwindow(*axes.getWindow())
                coord.setNDC(p0.x, p0.y)
                wcPick = coord.getWC(axes.viewport)
                curves = filter(lambda c: c.visible, axes.getCurves())
                for curve in curves:
                    for idx, x in enumerate(curve.x):
                        if x >= wcPick.x:
                            break
                    coord.setWC(x, curve.y[idx], axes.viewport)
                    points.append(coord.getNDC())
                    lstAxes.append(axes)
                    lstCurves.append(curve)
            if points:
                # calculate distance between p0 and point on curve
                norms = map(lambda p: (p0 - p).norm(), points)
                # nearest point
                idx = norms.index(min(norms))
                p = points[idx]
                axes = lstAxes[idx]
                curve = lstCurves[idx]
                coord.setNDC(p.x, p.y)
                coord.setWindow(*axes.getWindow())
            else:
                coord, axes, curve = None, None, None
        gr.setwindow(*window)
        return (coord, axes, curve)

    def select(self, p0, p1, width, height):
        window = gr.inqwindow()
        coord = CoordConverter(self._sizex, self._sizey)
        for axes in self._lstAxes:
            win = axes.getWindow()
            gr.setwindow(*win)
            p0World = coord.setNDC(p0.x, p0.y).getWC(self.viewport)
            p1World = coord.setNDC(p1.x, p1.y).getWC(self.viewport)
            xmin = min(p0World.x, p1World.x)
            xmax = max(p0World.x, p1World.x)
            ymin = min(p0World.y, p1World.y)
            ymax = max(p0World.y, p1World.y)
            axes.setWindow(xmin, xmax, ymin, ymax)
        gr.setwindow(*window)

    def pan(self, dp, width, height):
        window = gr.inqwindow()
        coord = CoordConverter(self._sizex, self._sizey)
        for axes in self._lstAxes:
            win = axes.getWindow()
            xmin, xmax, ymin, ymax = win
            gr.setwindow(*win)
            pmin = coord.setWC(xmin, ymin, axes.viewport, win).getNDC()
            pmax = coord.setWC(xmax, ymax, axes.viewport, win).getNDC()
            ndcWin = [pmin.x - dp.x, pmax.x - dp.x,
                      pmin.y - dp.y, pmax.y - dp.y]
            wmin = coord.setNDC(ndcWin[0], ndcWin[2]).getWC(axes.viewport)
            wmax = coord.setNDC(ndcWin[1], ndcWin[3]).getWC(axes.viewport)
            win = [wmin.x, wmax.x, wmin.y, wmax.y]
            axes.setWindow(*win)
        gr.setwindow(*window)

    def zoom(self, dpercent, p0=None, width=None, height=None):
        window = gr.inqwindow()
        coord = CoordConverter(self._sizex, self._sizey)
        for axes in self._lstAxes:
            win = axes.getWindow()
            xmin, xmax, ymin, ymax = win
            gr.setwindow(*win)
            # zoom from center
            pmin = coord.setWC(xmin, ymin, axes.viewport, win).getNDC()
            pmax = coord.setWC(xmax, ymax, axes.viewport, win).getNDC()
            winWidth = pmax.x - pmin.x
            winWidth_2 = winWidth / 2
            winHeight = pmax.y - pmin.y
            winHeight_2 = winHeight / 2
            dx_2 = winWidth_2 - (1 + dpercent) * winWidth_2
            dy_2 = winHeight_2 - (1 + dpercent) * winHeight_2
            ndcWin = [pmin.x - dx_2, pmax.x + dx_2,
                      pmin.y - dy_2, pmax.y + dy_2]
            # calculate ratio and adjust to ratio / zoom at origin p0
            if p0 is not None and width is not None and height is not None:
                ratio_xleft = (p0.x - pmin.x) / winWidth
                ratio_ydown = (p0.y - pmin.y) / winHeight
                # calculate new xmin, xmax, ymin, ymax with same ratio
                xmin_n = p0.x - ratio_xleft * (ndcWin[1] - ndcWin[0])
                xmax_n = xmin_n + (ndcWin[1] - ndcWin[0])
                ymin_n = p0.y - ratio_ydown * (ndcWin[3] - ndcWin[2])
                ymax_n = ymin_n + (ndcWin[3] - ndcWin[2])
                ndcWin = [xmin_n, xmax_n, ymin_n, ymax_n]
            wmin = coord.setNDC(ndcWin[0], ndcWin[2]).getWC(self.viewport)
            wmax = coord.setNDC(ndcWin[1], ndcWin[3]).getWC(self.viewport)
            win = [wmin.x, wmax.x, wmin.y, wmax.y]
            gr.setwindow(*window)
            if not axes.setWindow(*win):
                axes.setWindow(*window)
                self.reset()
                break

    def getROI(self, p0):
        res = None
        if self._legendROI:
            for roi in self._legendROI:
                if roi.isPointInside(p0):
                    res = roi
        return res

    def addAxes(self, *args, **kwargs):
        for plotAxes in args:
            if plotAxes and plotAxes not in self._lstAxes:
                plotAxes.viewport = self.viewport
                self._lstAxes.append(plotAxes)
                # overwrite ids (1: first axis, 2: second axis)
                self._countAxes += 1
                plotAxes._id = self._countAxes
        return self

    def drawGR(self):
        [xmin, xmax, ymin, ymax] = self.viewportscaled
         # redraw flag will be set if viewport has been adjusted in order to
         # display all legend items.
        redraw = False
        # -- draw title and subtitle -----------------------------------------
        charHeight = .027 * (ymax - ymin)
        charHeightUnscaled = .027 * (self.viewport[3] - self.viewport[2])
        if self.title or self.subTitle:
            dyTitle = 0
            dySubTitle = 0
            gr.settextalign(gr.TEXT_HALIGN_CENTER, gr.TEXT_VALIGN_BOTTOM)
            gr.setcharup(0., 1.)
            gr.setcharheight(charHeight)
            x = xmin + (xmax - xmin) / 2
            if self.title:
                dyTitle = charHeightUnscaled
            if self.title and self.subTitle:
                dySubTitle = charHeightUnscaled
            y = (self.viewport[3] + dyTitle + dySubTitle + charHeightUnscaled
                 + 0.01)

            if self.title:
                gr.text(x, y * self.sizey, self.title)
                y -= dyTitle + 0.01
            if self.subTitle:
                gr.text(x, y * self.sizey, self.subTitle)
        # -- draw axes and curves --------------------------------------------
        if self._lstAxes:
            for axes in self._lstAxes:
                axes.drawGR()
        dyXLabel = 0
        dxYLabel = 0
        # -- draw x- and y label ---------------------------------------------
        y = self.offsetXLabel + self.viewport[2]
        if self.xlabel:
            gr.settextalign(gr.TEXT_HALIGN_CENTER, gr.TEXT_VALIGN_TOP)
            gr.setcharup(0., 1.)
#            tby = gr.inqtext(0, 0, self.xlabel)[1]
#            tby = map(lambda y: gr.wctondc(0, y)[1], tby)
#            dyXLabel = max(tby) - min(tby) # already scaled
            dyXLabel = charHeightUnscaled
            y -= dyXLabel + charHeightUnscaled
            gr.text(xmin + (xmax - xmin) / 2., y * self.sizey, self.xlabel)
        if self.ylabel:
            gr.settextalign(gr.TEXT_HALIGN_CENTER, gr.TEXT_VALIGN_TOP)
            gr.setcharup(-1., 0.)
            tbx = gr.inqtext(0, 0, self.ylabel)[0]
            tbx = map(lambda y: gr.wctondc(0, y)[0], tbx)
            dxYLabel = max(tbx) - min(tbx)
            gr.text(self.offsetYLabel + xmin - dxYLabel / 2. - .075,
                    ymin + (ymax - ymin) / 2., self.ylabel)
            gr.setcharup(0., 1.)
        # -- draw legend and calculate ROIs for legend items -----------------
        if self._legend:
            x, y = xmin, y - dyXLabel - charHeightUnscaled
            if y < 0:
                vp = self.viewport
                vp[2] += dyXLabel
                self.viewport = vp # propagate update (calls setter)
                gr.clearws()
                self.drawGR()
            else:
                # preserve old values
                ltype = gr.inqlinetype()
                mtype = gr.inqmarkertype()
                lcolor = gr.inqlinecolorind()
                mcolor = gr.inqmarkercolorind()
                window = gr.inqwindow()
                scale = gr.inqscale()
                gr.setviewport(0, self.sizex, 0, self.sizey)
                gr.setwindow(0, self.sizex, 0, self.sizey)
                self._legendROI = []
                for axes in self._lstAxes:
                    if redraw:
                        break
                    for curve in axes.getCurves():
                        if curve.legend:
                            tbx, tby = gr.inqtext(0, 0, curve.legend)
                            textWidth = max(tbx) - min(tbx)
                            lineWidth = .1
                            if x + lineWidth + textWidth > self.sizex:
                                x = xmin
                                y -= 2 * charHeightUnscaled
                                if y < 0:
                                    vp = self.viewport
                                    vp[2] += 3 * charHeightUnscaled
                                    # propagate update (calls setter)
                                    self.viewport = vp
                                    redraw = True
                                    break
                            gr.setmarkertype(curve.markertype)
                            gr.setmarkercolorind(curve.markercolor)
                            if curve.linetype is not None:
                                ys = y * self.sizey # scaled y value
                                gr.setlinecolorind(curve.linecolor)
                                gr.setmarkercolorind(curve.markercolor)
                                gr.setlinetype(curve.linetype)
                                gr.polyline([x, x + lineWidth], [ys, ys])
                                if (curve.markertype != gr.MARKERTYPE_DOT
                                    and curve.markertype is not None):
                                    gr.setmarkertype(curve.markertype)
                                    gr.polymarker([x + .1 / 2.], [ys])
                            elif curve.markertype is not None:
                                gr.setmarkertype(curve.markertype)
                                gr.polymarker([x + .1 / 2.], [ys])

                            ybase = (y - charHeightUnscaled / 2) * self.sizey
                            ytop = (y + charHeightUnscaled / 2) * self.sizey
                            roi = RegionOfInterest(Point(x, ybase),
                                                   Point(x, ytop),
                                                   reference=curve,
                                           regionType=RegionOfInterest.LEGEND)
                            x += .11
                            if curve.visible:
                                gr.settextcolorind(1)
                            else:
                                gr.settextcolorind(83)
                            gr.settextalign(gr.TEXT_HALIGN_LEFT,
                                            gr.TEXT_VALIGN_HALF)
                            gr.text(x, ys, curve.legend)
                            gr.settextcolorind(1)
                            x += textWidth
                            roi.append(Point(x, ytop), Point(x, ybase))
                            self._legendROI.append(roi)
#                            gr.polyline(roi.x, roi.y)
                            tbx = gr.inqtext(0, 0, "X")[0]
                            charWidth = max(tbx) - min(tbx)
                            x += charWidth
                if not self._legendROI:
                    self._legendROI = None
                # restore old values
                gr.setlinecolorind(lcolor)
                gr.setmarkercolorind(mcolor)
                gr.setlinetype(ltype)
                gr.setmarkertype(mtype)
                # restore viewport and window
                gr.setviewport(*self.viewportscaled)
                gr.setwindow(*window)
                # restore scale
                gr.setscale(scale)
        if redraw:
            gr.clearws()
            self.drawGR()

class PlotCurve(GRDrawAttributes, GRMeta):

    COUNT = 0

    def __init__(self, x, y, errBar1=None, errBar2=None,
                 linetype=gr.LINETYPE_SOLID, markertype=gr.MARKERTYPE_DOT,
                 linecolor=None, markercolor=1, legend=None):
        super(PlotCurve, self).__init__(linetype, markertype, linecolor,
                                       markercolor)
        self._x, self._y = x, y
        self._e1, self._e2 = errBar1, errBar2
        self._linetype, self._markertype = linetype, markertype
        self._markercolor = markercolor
        self._legend = legend
        PlotCurve.COUNT += 1
        self._id = PlotCurve.COUNT
        if legend is None:
            self._legend = "curve %d" % self._id
        self._visible = True

    @property
    def errorBar1(self):
        """Get current ErrorBar #1 instance."""
        return self._e1

    @errorBar1.setter
    def errorBar1(self, value):
        self._e1 = value

    @property
    def errorBar2(self):
        """Get current ErrorBar #2 instance."""
        return self._e2

    @errorBar2.setter
    def errorBar2(self, value):
        self._e2 = value

    @property
    def x(self):
        """Get the current list/ndarray of x values."""
        return self._x

    @x.setter
    def x(self, lst):
        self._x = lst

    @property
    def y(self):
        """Get the current list/ndarray of y values."""
        return self._y

    @y.setter
    def y(self, lst):
        self._y = lst

    @property
    def legend(self):
        """Get current text for a legend."""
        return self._legend

    @legend.setter
    def legend(self, s):
        self._legend = s

    @property
    def visible(self):
        return self._visible

    @visible.setter
    def visible(self, bool):
        self._visible = bool

    def drawGR(self):
        if self.visible:
            # preserve old values
            ltype = gr.inqlinetype()
            mtype = gr.inqmarkertype()
            lcolor = gr.inqlinecolorind()
            mcolor = gr.inqmarkercolorind()

            if self.linetype is not None:
                gr.setlinecolorind(self.linecolor)
                gr.setmarkercolorind(self.markercolor)
                gr.setlinetype(self.linetype)
                gr.polyline(self.x, self.y)
                if (self.markertype != gr.MARKERTYPE_DOT and
                    self.markertype is not None):
                    gr.setmarkertype(self.markertype)
                    gr.polymarker(self.x, self.y)
            elif self.markertype is not None:
                gr.setmarkertype(self.markertype)
                gr.polymarker(self.x, self.y)
            if self.errorBar1:
                self.errorBar1.drawGR()
            if self.errorBar2:
                self.errorBar2.drawGR()
            # restore old values
            gr.setlinecolorind(lcolor)
            gr.setmarkercolorind(mcolor)
            gr.setlinetype(ltype)
            gr.setmarkertype(mtype)

class PlotAxes(GRViewPort, GRMeta):

    COUNT = 0

    def __init__(self, viewport, xtick=None, ytick=None, majorx=None,
                 majory=None, drawX=True, drawY=True):
        super(PlotAxes, self).__init__(viewport)
        self._xtick, self._ytick = xtick, ytick
        self.majorx, self.majory = majorx, majory
        self._drawX, self._drawY = drawX, drawY
        self._lstPlotCurve = []
        self._backgroundColor = 0
        self._window = None
        self._scale = 0
        self._grid = True
        self._resetWindow = False
        self._autoscale = False
        PlotAxes.COUNT += 1
        self._id = PlotAxes.COUNT
        self._xtick_callback, self._ytick_callback = None, None

    def getCurves(self):
        return self._lstPlotCurve

    def isXLogDomain(self):
        window = self.getWindow()
        if window:
            return DomainChecker.isInLogDomain(window[0], window[1])
        else:
            return False

    def isYLogDomain(self):
        window = self.getWindow()
        if window:
            return DomainChecker.isInLogDomain(window[2], window[3])
        else:
            return False

    @property
    def xtick(self):
        """get current user-defined interval between minor x ticks"""
        return self._xtick

    @xtick.setter
    def xtick(self, tickInterval):
        self._xtick = tickInterval

    @property
    def ytick(self):
        """get current user-defined interval between minor y ticks"""
        return self._ytick

    @ytick.setter
    def ytick(self, tickInterval):
        self._ytick = tickInterval

    @property
    def majorx(self):
        """get current user-defined number of minor tick intervals between
        major x tick marks"""
        return self._majorx

    @majorx.setter
    def majorx(self, minorCount):
        self._majorx = minorCount if minorCount > 0 or minorCount is None else 1

    @property
    def majory(self):
        """get current user-defined number of minor tick intervals between
        major y tick marks"""
        return self._majory

    @majory.setter
    def majory(self, minorCount):
        self._majory = minorCount if minorCount > 0 or minorCount is None else 1

    @property
    def scale(self):
        return self._scale

    @scale.setter
    def scale(self, options):
        self._scale = options

    @property
    def backgroundColor(self):
        return self._backgroundColor

    @backgroundColor.setter
    def backgroundColor(self, color):
        self._backgroundColor = color

    def isXDrawingEnabled(self):
        return self._drawX

    def setXDrawing(self, bool):
        self._drawX = bool

    def isYDrawingEnabled(self):
        return self._drawY

    def setYDrawing(self, bool):
        self._drawY = bool

    def setLogX(self, bool):
        if bool:
            self.scale |= gr.OPTION_X_LOG
        else:
            self.scale &= ~gr.OPTION_X_LOG

    def setLogY(self, bool):
        if bool:
            self.scale |= gr.OPTION_Y_LOG
        else:
            self.scale &= ~gr.OPTION_Y_LOG

    def setGrid(self, bool):
        self._grid = bool

    def isGridEnabled(self):
        return self._grid

    def setWindow(self, xmin, xmax, ymin, ymax):
        res = True
        # minimum window borders for log scale
        wmin, wmax = DomainChecker._ZERO * 10, DomainChecker._ZERO * 100
        # stay in log x domain
        if self.scale & gr.OPTION_X_LOG:
            if xmax < wmax:
                res = False
                xmax = wmax
            if xmin < wmin:
                res = False
                if self.autoscale:
                    xmin = 1 / 10 ** max(0, 3 - math.log10(xmax))
                else:
                    xmin = wmin
        # stay in log y domain
        if self.scale & gr.OPTION_Y_LOG:
            if ymax < wmax:
                res = False
                ymax = wmax
            if ymin < wmin:
                res = False
                if self.autoscale:
                    ymin = 1 / 10 ** max(0, 3 - math.log10(ymax))
                else:
                    ymin = wmin
        if DomainChecker.isInWindowDomain(xmin, xmax, ymin, ymax):
            self._window = [xmin, xmax, ymin, ymax]
        else:
            res = False
            _log.debug("xmin, xmax, ymin, ymax not in window domain.")
        return res

    def getWindow(self):
        if self._window:
            return list(self._window)
        else:
            return None

    def setXtickCallback(self, fp):
        self._xtick_callback = fp

    def setYtickCallback(self, fp):
        self._ytick_callback = fp

    def reset(self):
        self._resetWindow = True

    def isReset(self):
        return self._resetWindow or self.getWindow() is None

    @property
    def autoscale(self):
        return self._autoscale

    @autoscale.setter
    def autoscale(self, bool):
        self._autoscale = bool

    def getId(self):
        return self._id

    def drawGR(self):
        lstPlotCurve = self.getCurves()
        viewport = self.viewportscaled
        if lstPlotCurve:
            if self.isReset() or self.autoscale:
                # global xmin, xmax, ymin, ymax
                visibleCurves = filter(lambda curve: curve.visible,
                                       lstPlotCurve)
                if not visibleCurves:
                    visibleCurves = lstPlotCurve
                # calculate previous xmin, xmax
                xpmin = min(map(lambda curve: min(curve.x[:-1])
                                if len(curve.x) > 1 else curve.x[-1],
                                visibleCurves))
                xpmax = max(map(lambda curve: max(curve.x[:-1])
                                if len(curve.x) > 1 else curve.x[-1],
                                visibleCurves))
                # calculate xmin, xmax
                xl = [curve.x[-1] for curve in visibleCurves]
                yl = [curve.y[-1] for curve in visibleCurves]
                xmin = min(xpmin, *xl)
                xmax = max(xpmax, *xl)
                # calculate ymin, ymax
                ymin = min(map(lambda curve: min(curve.y),
                               visibleCurves))
                ymax = max(map(lambda curve: max(curve.y),
                               visibleCurves))
                # calculate ymin, ymax for last (added) point.
                ylmin, ylmax = min(yl), max(yl)

                if self.isReset():
                    self._resetWindow = False
                elif self.autoscale:
                    win = self.getWindow()
                    if win:
                        if xmin < xpmin:
                            # growing in negative direction
                            # hold max value
                            if xmax >= win[1]:
                                # win[1]: user max value
                                xmax = win[1]
                        elif xmax > xpmax:
                            # growing in positive direction
                            # hold min value
                            if xmin <= win[0]:
                                # win[0]: user min value
                                xmin = win[0]
                        ymin = ylmin if ylmin < win[2] else win[2]
                        ymax = ylmax if ylmax > win[3] else win[3]

                # take error bars into account
                # if error bar's center is in current window
                # adjust window margins to contain error bars.
                for curve in visibleCurves:
                    for bar in [curve.errorBar1, curve.errorBar2]:
                        if bar:
                            if bar.direction == ErrorBar.HORIZONTAL:
                                for i, x in enumerate(bar.x):
                                    if x >= xmin and x <= xmax:
                                        bneg, bpos = bar._dneg[i], bar._dpos[i]
                                        if bneg < xmin:
                                            xmin = bneg
                                        if bpos > xmax:
                                            xmax = bpos
                            elif bar.direction == ErrorBar.VERTICAL:
                                for i, y in enumerate(bar.y):
                                    if y >= ymin and y <= ymax:
                                        bneg, bpos = bar._dneg[i], bar._dpos[i]
                                        if bneg < ymin:
                                            ymin = bneg
                                        if bpos > ymax:
                                            ymax = bpos

                if ((not self.autoscale
                     or math.fabs(xmax - xmin) < DomainChecker._EPSILON)
                    and self.scale & gr.OPTION_X_LOG == 0):
                    xmin, xmax = gr.adjustrange(xmin, xmax)
                if ((not self.autoscale or
                     math.fabs(ymax - ymin) < DomainChecker._EPSILON)
                    and self.scale & gr.OPTION_Y_LOG == 0):
                    ymin, ymax = gr.adjustrange(ymin, ymax)

                window = [xmin, xmax, ymin, ymax]
                self.setWindow(*window)
            else:
                window = self.getWindow()
                xmin, xmax, ymin, ymax = window

            if (window[0] > xmin or xmin > window[1] or window[2] > ymin or
                ymin > window[3]):
                #GKS: Rectangle definition is invalid in routine SET_WINDOW
                #origin outside current window
                self.reset()
                self.drawGR()
            else:
                if self.scale & gr.OPTION_X_LOG:
                    xtick = majorx = 1
                else:
                    majorx = self._majorx if self._majorx is not None else 5
                    if self._xtick is not None:
                        xtick = self._xtick
                    else:
                        xtick = gr.tick(xmin, xmax) / majorx
                if self.scale & gr.OPTION_Y_LOG:
                    ytick = majory = 1
                else:
                    majory = self._majory if self._majory is not None else 5
                    if self._ytick is not None:
                        ytick = self._ytick
                    else:
                        ytick = gr.tick(ymin, ymax) / majory
                # window may has been modified - get most recent window
                window = self.getWindow()
                xmin, xmax, ymin, ymax = window
                gr.setviewport(*self.viewportscaled)
                gr.setwindow(*window)
                gr.setscale(self.scale)
                if self.backgroundColor and self.getId() == 1:
                    gr.setfillintstyle(1)
                    gr.setfillcolorind(self.backgroundColor)
                    gr.fillrect(*window)
                charHeight = .024 * (viewport[3] - viewport[2])
                gr.setcharheight(charHeight)
                if self.isGridEnabled() and self.getId() == 1:
                    # delta majorx, delta majory
                    dmx = xtick * majorx
                    dmy = ytick * majory
                    # xorg, yorg
                    i = int(xmin / dmx) if xmin < 0 else int(xmin / dmx + 1)
                    j = int(ymin / dmy) if ymin < 0 else int(ymin / dmy + 1)
                    xorg = i * dmx
                    yorg = j * dmy
                    gr.grid(xtick, ytick, xorg, yorg, majorx, majory)
                if self.getId() == 1:
                    # first x, y axis
                    if not self.isXDrawingEnabled():
                        majorx = -majorx
                    if not self.isYDrawingEnabled():
                        majory = -majory
                    gr.axeslbl(xtick, ytick, xmin, ymin, majorx, majory,
                               0.01, self._xtick_callback, self._ytick_callback)
                elif self.getId() == 2:
                    # second x, y axis
                    if not self.isXDrawingEnabled():
                        majorx = -majorx
                    if not self.isYDrawingEnabled():
                        majory = -majory
                    gr.axeslbl(xtick, ytick, xmax, ymax, majorx, majory,
                               - 0.01, self._xtick_callback,
                               self._ytick_callback)
                for curve in lstPlotCurve:
                    curve.drawGR()

    def plot(self, *args, **kwargs):
        if len(args) > 0 and len(args) % 2 == 0:
            self._lstPlotCurve = []
            for i in range(0, len(args) / 2):
                x = args[2 * i]
                y = args[2 * i + 1]
                self._lstPlotCurve.append(PlotCurve(x, y))
        return self

    def addCurves(self, *args, **kwargs):
        for plotCurve in args:
            if plotCurve and plotCurve not in self._lstPlotCurve:
                self._lstPlotCurve.append(plotCurve)
        return self

def ndctowc(x, y):
    try:
        iter(x)
        iter(y)
        resX = []
        resY = []
        for xi, yi in map(lambda x, y: gr.ndctowc(x, y), x, y):
            resX.append(xi)
            resY.append(yi)
        return resX, resY
    except TypeError:
        return gr.ndctowc(x, y)

def wctondc(x, y):
    try:
        iter(x)
        iter(y)
        resX = []
        resY = []
        for xi, yi in map(lambda x, y: gr.wctondc(x, y), x, y):
            resX.append(xi)
            resY.append(yi)
        return resX, resY
    except TypeError:
        return gr.wctondc(x, y)

def _guessdimension(len):
    x = int(math.sqrt(len))
    d = []
    while x >= 1:
        y = len / x
        if x * y == len:
            d.append((x, y))
        x -= 1
    return sorted(d, key=lambda t: t[1] - t[0])

def readfile(path, separator=''):
    fp = open(path, "r")
    if separator != '':
        while True:
            line = fp.readline()
            if not line:
                break
            if line[0] == separator:
                break
    data = []
    for line in fp.readlines():
        for item in line.split():
            data.append(float(item))
    fp.close()
    return data

def plot(x, y,
         bgcolor=0,
         viewport=(0.1, 0.95, 0.1, 0.95),
         window=None,
         scale=0,
         grid=True,
         linetype=gr.LINETYPE_SOLID,
         markertype=gr.MARKERTYPE_DOT,
         clear=True,
         update=True):
    if clear:
        gr.clearws()
    if window == None:
        if scale & gr.OPTION_X_LOG == 0:
            xmin, xmax = gr.adjustrange(min(x), max(x))
        else:
            xmin, xmax = (min(x), max(x))
        if scale & gr.OPTION_Y_LOG == 0:
            ymin, ymax = gr.adjustrange(min(y), max(y))
        else:
            ymin, ymax = (min(y), max(y))
    else:
        xmin, xmax, ymin, ymax = window
    if scale & gr.OPTION_X_LOG == 0:
        majorx = 5
        xtick = gr.tick(xmin, xmax) / majorx
    else:
        xtick = majorx = 1
    if scale & gr.OPTION_Y_LOG == 0:
        majory = 5
        ytick = gr.tick(ymin, ymax) / majory
    else:
        ytick = majory = 1
    gr.setviewport(viewport[0], viewport[1], viewport[2], viewport[3])
    gr.setwindow(xmin, xmax, ymin, ymax)
    gr.setscale(scale)
    if bgcolor != 0:
        gr.setfillintstyle(1)
        gr.setfillcolorind(bgcolor)
        gr.fillrect(xmin, xmax, ymin, ymax)
    charheight = 0.024 * (viewport[3] - viewport[2])
    gr.setcharheight(charheight)
    if grid:
         gr.grid(xtick, ytick, xmax, ymax, majorx, majory)
    gr.axes(xtick, ytick, xmin, ymin, majorx, majory, 0.01)
    gr.axes(xtick, ytick, xmax, ymax, -majorx, -majory, -0.01)
    gr.setlinetype(linetype)
    gr.polyline(x, y)
    if markertype != gr.MARKERTYPE_DOT:
        gr.setmarkertype(markertype)
        gr.polymarker(x, y)
    if update:
        gr.updatews()

def plot3d(z,
           viewport=(0.1, 0.9, 0.1, 0.9),
           rotation=30,
           tilt=70,
           colormap=1,
           option=4,
           contours=True,
           xtitle='',
           ytitle='',
           ztitle='',
           accelerate=False):
    gr.clearws()
    xmin, ymin = (1, 1)
    if type(z) == ndarray:
        xmax, ymax = z.shape
        zmin = z.min()
        zmax = z.max()
        z = z.ravel()
    else:
        xmax, ymax = _guessdimension(len(z))[0]
        zmin = min(z)
        zmax = max(z)
    xtick = gr.tick(xmin, xmax) / 5
    ytick = gr.tick(ymin, ymax) / 5
    x = range(1, xmax + 1)
    y = range(1, ymax + 1)
    zmin, zmax = gr.adjustrange(zmin, zmax)
    ztick = gr.tick(zmin, zmax) / 5
    gr.setviewport(viewport[0], viewport[1], viewport[2], viewport[3])
    gr.setwindow(xmin, xmax, ymin, ymax)
    gr.setspace(zmin, zmax, rotation, tilt)
    charheight = 0.024 * (viewport[3] - viewport[2])
    gr.setcharheight(charheight)
    gr.setcolormap(colormap)
    if accelerate:
        gr3.surface(x, y, z, option)
    else:
        gr.surface(x, y, z, option)

    if rotation != 0 or tilt != 90:
        gr.axes3d(xtick, 0, ztick, xmin, ymin, zmin, 5, 0, 5, -0.01)
        gr.axes3d(0, ytick, 0, xmax, ymin, zmin, 0, 5, 0, 0.01)
    if contours:
        gr.contour(x, y, range(1), z, 0)
    if rotation == 0 and tilt == 90:
        gr.axes(xtick, ytick, xmin, ymin, 5, 5, -0.01)
    if xtitle != '' or ytitle != '' or ztitle != '':
        gr.titles3d(xtitle, ytitle, ztitle)
    gr.updatews()

def imshow(data, cmap=gr.COLORMAP_GRAYSCALE):
    height, width = data.shape
    d = data.reshape(width * height).astype(float)
    d = 8 + 72 * (d - min(d)) / (max(d) - min(d))
    ca = d.astype(int)
    gr.clearws()
    if width < height:
        ratio = float(width) / height
        xmin = max(0.5 * (1 - ratio), 0)
        xmax = min(xmin + ratio, 1)
        ymin = 0
        ymax = 1
    else:
        ratio = float(height) / width
        xmin = 0
        xmax = 1
        ymin = max(0.5 * (1 - ratio), 0)
        ymax = min(ymin + ratio, 1)
    gr.selntran(0)
    gr.setcolormap(cmap)
    gr.cellarray(xmin, xmax, ymin, ymax, width, height, ca)
    gr.updatews()

def delay(seconds):
    time.sleep(seconds)



import sys
from math import sqrt, pi, atan2
import zlib
import struct
import gks
from constants import *

fonts = [
    "Times New Roman", "Arial", "Courier", "Open Symbol",
    "Bookman Old Style", "Century Schoolbook", "Century Gothic",
    "Book Antiqua"
]

capheights = [
    0.662, 0.660, 0.681, 0.662,
    0.729, 0.729, 0.729, 0.729,
    0.583, 0.583, 0.583, 0.583,
    0.667,
    0.681, 0.681, 0.681, 0.681,
    0.722, 0.722, 0.722, 0.722,
    0.739, 0.739, 0.739, 0.739,
    0.694, 0.693, 0.683, 0.683
]

font_map = [
    22, 9, 5, 14, 18, 26, 13, 1,
    24, 11, 7, 16, 20, 28, 13, 3,
    23, 10, 6, 15, 19, 27, 13, 2,
    25, 12, 8, 17, 21, 29, 13, 4
]

predef_styli = [1, 1, 1, 2, 3]
predef_ints = [0, 1, 3, 3, 3]

a = [1.0 for i in range(MAX_TNR)]
b = [0.0 for i in range(MAX_TNR)]
c = [1.0 for i in range(MAX_TNR)]
d = [0.0 for i in range(MAX_TNR)]


class ws_state_list(object):
    def __init__(self):
        self.state = None
        self.a = None
        self.b = None
        self.c = None
        self.d = None
        self.width = None
        self.height = None
        self.window = None
        self.viewport = None
        self.rgb = None
        self.strokeStyle = None
        self.fillStyle = None
        self.lineWidth = None
        self.dashes = None
        self.valign = None
        self.halign = None
        self.font = None
        self.rect = None
        self.clip_rect = None


class Html_output(object):
    def __init__(self, gks_state_list, data, n):
        if n >= 250:
            print("Maximum number of HTML output files reached")
            sys.exit(0)
        self.filename = 'gks_p%03d.html' % (n)
        self.title = 'GKS'
        self.gkss = gks_state_list
        self.p = ws_state_list()
        self.file = open(self.filename, 'w')
        self.p.width = 500
        self.p.height = 500
        self.p.window = [0.0, 1.0, 0.0, 1.0]
        self.p.viewport = [self.p.width * MWIDTH / WIDTH, 0.,
                           self.p.height * MHEIGHT / HEIGHT]
        self.p.rgb = [[] for i in range(MAX_COLOR)]
        self.p.dashes = []
        self.indentation = 0
        self.p.rect = [() for i in range(MAX_TNR)]
        self.p.clip_rect = ()
        self.transparency = 1.0

        self.set_xform()
        self.init_norm_xform()
        self.init_colors()

        if not self.file:
            print("Can't open HTML output file")

        self.write('<!DOCTYPE html>\n')
        self.write('<html>\n')
        self.write('  <head>\n')
        self.write('    <meta charset="utf-8" />\n')
        self.write('    <title>{0}</title>\n'.format(self.title))
        self.write('  </head>\n')
        self.write('  <script src="modernizr.js"></script>\n')
        self.write('  <body>\n')
        self.write(
            '    <canvas id="html-canvas" width="{0}" height="{1}" tabindex="1" style="outline-style:none;"></canvas>\n'.format(
                self.p.width, self.p.height))
        self.write('\n')
        self.write('<script src="gks.js"></script>\n')
        self.write('<script type="text/javascript">\n')
        self.indentation = 1
        self.init_canvas()

        gks.set_fill_callback(self.fill_routine)
        gks.set_line_callback(self.line_routine)

        self.write('base_width = canvas.width;\n')
        self.write('base_height = canvas.height;\n')

        self.write('function draw() {\n')

        self.footer = '''  }
  draw();
</script>
<div id="buttons"></div>
<script>
    if (Modernizr.touch) {
        var container = document.getElementById("buttons");
            container.innerHTML = "<table border='0'><tr><td><button onclick='zoom_in();'>Zoom In</button></td><td><button onclick='zoom_out();'>Zoom Out</button></td><td><button onclick='reset_zoom();'>Reset</button></td></tr></table>";
        }
</script>
</body>
</html>
'''
        for d in data:
            getattr(self, d[0])(*d[1])

        self.file.write(self.footer)
        self.file.close()

    def text(self, xst, yst, n, text):

        tx_font = self.gkss.txfont if self.gkss.asf[6] else predef_font[self.gkss.tindex - 1]
        tx_prec = self.gkss.txprec if self.gkss.asf[6] else predef_prec[self.gkss.tindex - 1]
        tx_color = self.gkss.txcoli if self.gkss.asf[9] else 1
        color = self.p.rgb[tx_color]
        color.append(self.transparency)

        if self.gkss.version > 4:
            ln_width = round(self.p.height / 500.0)
            if ln_width < 1:
                ln_width = 1
        else:
            ln_width = 1
        if ln_width < 1:
            ln_width = 1
        if self.p.fillStyle != color:
            self.p.fillStyle = color
            self.write('c.fillStyle="rgba({0},{1},{2},{3})";\n'.format(*color))
        if self.p.strokeStyle != color:
            self.p.strokeStyle = color
            self.write('c.strokeStyle="rgba({0},{1},{2},{3})";\n'.format(*color))
        if tx_prec == GKS_K_TEXT_PRECISION_STRING:
            self.set_font(tx_font)
            (x, y) = self.WC_to_NDC(xst, yst, self.gkss.cntnr)
            (x, y) = self.seg_xform(x, y)
            self.text_routine(x, y, n, text)
        else:
            gks.emul_text(xst, yst, text)

    def text_routine(self, x, y, nchars, text):
        (xs, ys) = self.NDC_to_DC(x, y)

        halign = self.gkss.txal[0]
        valign = self.gkss.txal[1]

        if self.p.valign != valign:
            self.p.valign = valign
            if valign == GKS_K_TEXT_VALIGN_NORMAL:
                self.write('c.textBaseline="alphabetic";\n')
            elif valign == GKS_K_TEXT_VALIGN_TOP:
                self.write('c.textBaseline="top";\n')
            elif valign == GKS_K_TEXT_VALIGN_CAP:
                self.write('c.textBaseline="hanging";\n')
            elif valign == GKS_K_TEXT_VALIGN_HALF:
                self.write('c.textBaseline="middle";\n')
            elif valign == GKS_K_TEXT_VALIGN_BASE:
                self.write('c.textBaseline="alphabetic";\n')
            elif valign == GKS_K_TEXT_VALIGN_BOTTOM:
                self.write('c.textBaseline="bottom";\n')

        if self.p.halign != halign:
            if halign == GKS_K_TEXT_HALIGN_NORMAL:
                self.write('c.textAlign="left";\n')
            elif halign == GKS_K_TEXT_HALIGN_LEFT:
                self.write('c.textAlign="left";\n')
            elif halign == GKS_K_TEXT_HALIGN_CENTER:
                self.write('c.textAlign="center";\n')
            elif halign == GKS_K_TEXT_HALIGN_RIGHT:
                self.write('c.textAlign="right";\n')

        if self.p.alpha > 0:
            self.write('c.save();\n')
            self.write('c.translate({0}, {1});\n'.format(xs, ys))
            self.write('c.rotate({0});\n'.format(-self.p.alpha))
            self.write('c.fillText("{0}", 0, 0);\n'.format(text))
            self.write('c.restore();\n')
        else:
            self.write('c.fillText("{0}", {1}, {2});\n'.format(text, xs, ys))

    def cellarray(self, xmin, xmax, ymin, ymax, dx, dy, dimx, colia):
        self.image_routine(xmin, xmax, ymin, ymax, dx, dy, dimx, colia, False)

    def draw_image(self, xmin, xmax, ymin, ymax, dx, dy, dimx, colia):
        self.image_routine(xmin, xmax, ymin, ymax, dx, dy, dimx, colia, True)

    def write_png(self, buf, width, height):
        width_byte_4 = width * 4
        raw_data = b"".join(b'\x00' + bytes(bytearray(buf[span:span + width_byte_4])) for span in range(0, height * width * 4, width_byte_4))
        def png_pack(png_tag, data):
            chunk_head = png_tag + data
            return struct.pack("!I", len(data)) + chunk_head + struct.pack("!I", 0xFFFFFFFF & zlib.crc32(chunk_head))
        return b"".join([
            b'\x89PNG\r\n\x1a\n',
            png_pack(b'IHDR', struct.pack("!2I5B", width, height, 8, 6, 0, 0, 0)),
            png_pack(b'IDAT', zlib.compress(raw_data, 9)),
            png_pack(b'IEND', b'')])

    def image_routine(self, xmin, xmax, ymin, ymax, dx, dy, dimx, colia, true_color):
        (x1, y1) = self.WC_to_NDC(xmin, ymax, self.gkss.cntnr)
        (x1, y1) = self.seg_xform(x1, y1)
        (ix1, iy1) = self.NDC_to_DC(x1, y1)

        (x2, y2) = self.WC_to_NDC(xmax, ymin, self.gkss.cntnr)
        (x2, y2) = self.seg_xform(x2, y2)
        (ix2, iy2) = self.NDC_to_DC(x2, y2)

        pix_buf = []

        ix1 = int(ix1)
        iy1 = int(iy1)
        ix2 = int(ix2)
        iy2 = int(iy2)

        width = abs(ix2 - ix1) + 1
        height = abs(iy2 - iy1) + 1

        x = min(ix1, ix2)
        y = min(iy1, iy2)

        swapx = ix1 > ix2
        swapy = iy1 < iy2

        for j in range(height):
            iy = dy * j / height
            if swapy:
                iy = dy - 1 - iy
            for i in range(width):
                ix = dx * i / width
                if swapx:
                    ix = dx - 1 - ix
                if not true_color:
                    ci = colia[iy * dimx + ix]
                    (red, green, blue) = self.p.rgb[ci][:3]
                    alpha = int(self.transparency * 255)
                else:
                    rgb = colia[iy * dimx + ix]
                    red = (rgb & 0xff)
                    green = (rgb & 0xff00) >> 8
                    blue = (rgb & 0xff0000) >> 16
                    alpha = (rgb & 0xff000000) >> 24

                pix_buf.append(red)
                pix_buf.append(green)
                pix_buf.append(blue)
                pix_buf.append(alpha)

        png = self.write_png(pix_buf, width, height)

        import base64
        enc_png = base64.b64encode(png)
        data_uri = 'data:image/png;base64, {0}'.format(enc_png)
        self.write('var imageObj = new Image();\n')
        self.write('imageObj.onload = function() {\n')
        self.indentation += 1
        self.write('c.drawImage(this, {0}, {1});\n'.format(x, y))

        self.footer = (self.indentation - 1) * '  ' + 'imageObj.src = "{0}";\n'.format(data_uri) + self.footer
        self.footer = (self.indentation - 1) * '  ' + '};\n' + self.footer

    def fillarea(self, n, px, py):
        fl_color = self.gkss.facoli if self.gkss.asf[12] else 1
        ln_width = max(1, round(self.p.height / 500.0)) if self.gkss.version > 4 else 1
        if self.p.lineWidth != ln_width:
            self.p.lineWidth = ln_width
            self.write('c.lineWidth = {0};\n'.format(ln_width))
        color = self.p.rgb[fl_color]
        color.append(self.transparency)
        if self.p.fillStyle != color:
            self.p.fillStyle = color
            self.write('c.fillStyle="rgba({0},{1},{2},{3})";\n'.format(*color))
        self.fill_routine(n, px, py, self.gkss.cntnr)

    def fill_routine(self, n, px, py, tnr):
        self.write('c.beginPath();\n')

        if self.p.dashes != []:
            self.p.dashes = []
            self.write('set_dashes(c, []);\n')

        (xn, yn) = self.WC_to_NDC(px[0], py[0], self.gkss.cntnr)
        (xn, yn) = self.seg_xform(xn, yn)
        (xd, yd) = self.NDC_to_DC(xn, yn)
        self.write('c.moveTo({0}, {1});\n'.format(xd, yd))
        for i in range(1, n):
            (xn, yn) = self.WC_to_NDC(px[i], py[i], self.gkss.cntnr)
            (xn, yn) = self.seg_xform(xn, yn)
            (xd, yd) = self.NDC_to_DC(xn, yn)
            self.write('c.lineTo({0}, {1});\n'.format(xd, yd))
        self.write('c.closePath();\n')

        fl_inter = self.gkss.ints if self.gkss.asf[10] else predef_ints[self.gkss.findex - 1]
        if fl_inter == GKS_K_INTSTYLE_PATTERN or fl_inter == GKS_K_INTSTYLE_HATCH:
            fl_style = self.gkss.styli if self.gkss.asf[11] else predef_styli[self.gkss.findex - 1]
            if fl_inter == GKS_K_INTSTYLE_HATCH:
                fl_style += HATCH_STYLE
            if fl_style >= PATTERNS:
                fl_style = 1
            pattern = gks.inq_pattern_array(fl_style)
            self.write('var pcan = document.createElement("canvas");\n')
            self.write('pcan.width = 8;\n')
            self.write('pcan.height = {0};\n'.format(len(pattern)))
            self.write('var pctx = pcan.getContext("2d");\n')
            if self.p.fillStyle != [0, 0, 0]:
                self.write('c.fillStyle="rgba({0},{1},{2},{3})";\n'.format(*self.p.fillStyle))
            for j in range(len(pattern)):
                for i in range(8):
                    a = (1 << i) & pattern[j]
                    if not a:
                        self.write(
                            'pctx.rect({0}, {1}, 1, 1);\n'.format((i + 7) % 8, (j + (len(pattern) - 1)) % len(pattern)))
            self.write('pctx.fill();\n')
            self.write('var pattern = c.createPattern(pcan, "repeat");\n')
            self.write('c.fillStyle = pattern;\n')
            self.write('c.fill();\n')
        elif fl_inter == GKS_K_INTSTYLE_SOLID:
            self.write('c.fill();\n')
        else:
            self.write('c.stroke();\n')

    def set_window(self, tnr, xmin, xmax, ymin, ymax):
        self.gkss.window[tnr][0] = xmin
        self.gkss.window[tnr][1] = xmax
        self.gkss.window[tnr][2] = ymin
        self.gkss.window[tnr][3] = ymax

        self.set_xform()
        self.set_norm_xform(tnr, self.gkss.window[tnr], self.gkss.viewport[tnr])

    def set_viewport(self, tnr, xmin, xmax, ymin, ymax):
        self.gkss.viewport[tnr][0] = xmin
        self.gkss.viewport[tnr][1] = xmax
        self.gkss.viewport[tnr][2] = ymin
        self.gkss.viewport[tnr][3] = ymax

        self.set_xform()
        self.set_norm_xform(tnr, self.gkss.window[tnr], self.gkss.viewport[tnr])
        if tnr == self.gkss.cntnr:
            self.set_clip_rect(tnr)

    def set_asf(self, *asf):
        for i in range(13):
            self.gkss.asf[i] = asf[i]

    def polyline(self, n, px, py):
        ln_width = self.gkss.lwidth if self.gkss.asf[1] else 1
        ln_type = self.gkss.ltype if self.gkss.asf[0] else self.gkss.lindex
        ln_color = self.gkss.plcoli if self.gkss.asf[2] else 1

        if self.gkss.version > 4:
            ln_width *= self.p.height / 500.0
        if ln_color <= 0 or ln_color >= MAX_COLOR:
            ln_color = 1
        ln_width = max(1, round(ln_width))

        color = self.p.rgb[ln_color]
        color.append(self.transparency)
        if self.p.strokeStyle != color:
            self.p.strokeStyle = color
            self.write('c.strokeStyle="rgba({0},{1},{2},{3})";\n'.format(*color))
        if self.p.lineWidth != ln_width:
            self.p.lineWidth = ln_width
            self.write('c.lineWidth = {0};\n'.format(ln_width))

        self.write('c.beginPath();\n')
        dashes = gks.get_dash_list(ln_type, ln_width)
        if self.p.dashes != dashes:
            self.p.dashes = dashes
            self.write('set_dashes(c, {0});\n'.format(str(dashes)))

        (xn, yn) = self.WC_to_NDC(px[0], py[0], self.gkss.cntnr)
        (xn, yn) = self.seg_xform(xn, yn)
        (xd, yd) = self.NDC_to_DC(xn, yn)
        self.write('c.moveTo({0}, {1});\n'.format(xd, yd))
        for i in range(1, n):
            (xn, yn) = self.WC_to_NDC(px[i], py[i], self.gkss.cntnr)
            (xn, yn) = self.seg_xform(xn, yn)
            (xd, yd) = self.NDC_to_DC(xn, yn)
            self.write('c.lineTo({0}, {1});\n'.format(xd, yd))
        self.write('c.stroke();\n')

    def line_routine(self, n, px, py, ltype, tnr):
        (xn, yn) = self.WC_to_NDC(px[0], py[0], self.gkss.cntnr)
        (xn, yn) = self.seg_xform(xn, yn)
        (xd, yd) = self.NDC_to_DC(xn, yn)
        self.write('c.beginPath();\n')
        self.write('c.moveTo({0}, {1});\n'.format(xd, yd))
        for i in range(1, n):
            (xn, yn) = self.WC_to_NDC(px[i], py[i], self.gkss.cntnr)
            (xn, yn) = self.seg_xform(xn, yn)
            (xd, yd) = self.NDC_to_DC(xn, yn)
            self.write('c.lineTo({0}, {1});\n'.format(xd, yd))
        self.write('c.stroke();\n')

    def init_canvas(self):
        self.write('var canvas=document.getElementById("html-canvas");\n')
        self.write('var c=canvas.getContext("2d");\n')

    def draw_point(self, x, y):
        self.write('c.fillRect({0}, {1}, 1, 1);\n'.format(x, y))

    def draw_line(self, x1, y1, x2, y2):
        self.write('c.beginPath();\n')
        self.write('c.moveTo({0}, {1});\n'.format(x1, y1))
        self.write('c.lineTo({0}, {1});\n'.format(x2, y2))
        self.write('c.stroke();\n')

    def draw_marker(self, xn, yn, mtype, mscale, mcolor):
        marker = [
            [5, 9, -4, 7, 4, 7, 7, 4, 7, -4,  # omark
             4, -7, -4, -7, -7, -4, -7, 4,
             -4, 7, 3, 9, -4, 7, 4, 7, 7, 4,
             7, -4, 4, -7, -4, -7, -7, -4,
             -7, 4, -4, 7, 0],
            [5, 13, -2, 8, 2, 8, 2, 2, 8, 2,  # hollow plus
             8, -2, 2, -2, 2, -8, -2, -8,
             -2, -2, -8, -2, -8, 2, -2, 2,
             -2, 8, 3, 13, -2, 8, 2, 8,
             2, 2, 8, 2, 8, -2, 2, -2, 2, -8,
             -2, -8, -2, -2, -8, -2, -8, 2,
             -2, 2, -2, 8, 0],
            [4, 4, -8, 0, 4, 7, 4, -7,  # solid triangle right
             -8, 0, 0],
            [4, 4, 8, 0, -4, -7, -4, 7,  # solid triangle left
             8, 0, 0],
            [5, 4, 0, 8, 7, -4, -7, -4, 0, 8,  # triangle up down
             5, 4, 0, -8, -7, 4, 7, 4, 0, -8,
             3, 4, 0, 8, 7, -4, -7, -4, 0, 8,
             3, 4, 0, -8, -7, 4, 7, 4, 0, -8,
             0],
            [4, 11, 0, 9, 2, 2, 9, 3, 3, -1,  # solid star
             6, -8, 0, -3, -6, -8, -3, -1,
             -9, 3, -2, 2, 0, 9, 0],
            [5, 11, 0, 9, 2, 2, 9, 3, 3, -1,  # hollow star
             6, -8, 0, -3, -6, -8, -3, -1,
             -9, 3, -2, 2, 0, 9,
             3, 11, 0, 9, 2, 2, 9, 3, 3, -1,
             6, -8, 0, -3, -6, -8, -3, -1,
             -9, 3, -2, 2, 0, 9, 0],
            [4, 5, 0, 9, 9, 0, 0, -9, -9, 0,  # solid diamond
             0, 9, 0],
            [5, 5, 0, 9, 9, 0, 0, -9, -9, 0,  # hollow diamond
             0, 9, 3, 5, 0, 9, 9, 0, 0, -9,
             -9, 0, 0, 9, 0],
            [4, 5, 9, 9, -9, -9, 9, -9, -9, 9,  # solid hourglass
             9, 9, 0],
            [5, 5, 9, 9, -9, -9, 9, -9, -9, 9,  # hollow hourglass
             9, 9, 3, 5, 9, 9, -9, -9, 9, -9,
             -9, 9, 9, 9, 0],
            [4, 5, 9, 9, 9, -9, -9, 9, -9, -9,  # solid bowtie
             9, 9, 0],
            [5, 5, 9, 9, 9, -9, -9, 9, -9, -9,  # hollow bowtie
             9, 9, 3, 5, 9, 9, 9, -9, -9, 9,
             -9, -9, 9, 9, 0],
            [4, 5, 9, 9, 9, -9, -9, -9, -9, 9,  # solid square
             9, 9, 0],
            [5, 5, 9, 9, 9, -9, -9, -9, -9, 9,  # hollow square
             9, 9, 3, 5, 9, 9, 9, -9, -9, -9,
             -9, 9, 9, 9, 0],
            [4, 4, -9, 9, 9, 9, 0, -9, -9, 9,  # solid triangle down
             0],
            [5, 4, -9, 9, 9, 9, 0, -9, -9, 9,  # hollow triangle down
             3, 4, -9, 9, 9, 9, 0, -9, -9, 9,
             0],
            [4, 4, 0, 9, 9, -9, -9, -9, 0, 9,  # solid triangle up
             0],
            [5, 4, 0, 9, 9, -9, -9, -9, 0, 9,  # hollow triangle up
             3, 4, 0, 9, 9, -9, -9, -9, 0, 9, 0],
            [7, 0, 360, 0],  # solid circle
            [0],  # not used
            [1, 0],  # dot
            [2, 0, 0, 0, 9, 2, 0, 0, 9, 0,  # plus
             2, 0, 0, 0, -9, 2, 0, 0, -9, 0,
             0],
            [2, 0, 0, 0, 9, 2, 0, 0, 9, 3,  # asterisk
             2, 0, 0, 6, -9, 2, 0, 0, -6, -9,
             2, 0, 0, -9, 3, 0],
            [8, 0, 360, 6, 0, 360, 0],  # circle
            [2, 0, 0, 9, 9, 2, 0, 0, 9, -9,  # diagonal cross
             2, 0, 0, -9, -9, 2, 0, 0, -9, 9, 0]]

        if self.gkss.version > 4:
            mscale *= self.p.height / 500.0
        r = int(3 * mscale)
        scale = mscale / 3.0

        xr = r
        yr = 0
        (xr, yr) = self.seg_xform_rel(xr, yr)
        r = round(sqrt(xr * xr + yr * yr))

        (x, y) = self.NDC_to_DC(xn, yn)

        pc = 0
        mtype = mtype + 20 if (2 * r > 1) else 21
        while (True):
            op = marker[mtype][pc]
            if op == 1:  # point
                self.draw_point(x, y)
            elif op == 2:  # line
                x1 = scale * marker[mtype][pc + 1]
                y1 = scale * marker[mtype][pc + 2]
                (x1, y1) = self.seg_xform_rel(x1, y1)

                x2 = scale * marker[mtype][pc + 2 + 1]
                y2 = scale * marker[mtype][pc + 2 + 2]
                (x2, y2) = self.seg_xform_rel(x2, y2)

                self.draw_line(x - x1, y - y1, x - x2, y - y2)
                pc += 4

            elif op in (3, 4, 5):  # 3 - polygon   4 - filled polygon   5 - hollow polygon
                xr = scale * marker[mtype][pc + 2]
                yr = scale * marker[mtype][pc + 3]
                (xr, yr) = self.seg_xform_rel(xr, yr)

                self.write('c.beginPath();\n')
                self.write('c.moveTo({0}, {1});\n'.format(x - xr, y - yr))

                for i in range(1, marker[mtype][pc + 1]):
                    xr = scale * marker[mtype][pc + 2 + 2 * i]
                    yr = scale * marker[mtype][pc + 3 + 2 * i]

                    (xr, yr) = self.seg_xform_rel(xr, yr)
                    self.write('c.lineTo({0}, {1});\n'.format(x - xr, y - yr))

                self.write('c.closePath();\n')

                if not op == 3:
                    if op == 5:
                        self.write('c.fillStyle = "#FFFFFF";\n')
                    else:
                        self.write('c.fillStyle = "#000000";\n')
                    self.write('c.fill();\n')
                    self.write('c.fillStyle = "#000000";\n')
                else:
                    self.write('c.stroke();\n')

                pc += 1 + 2 * marker[mtype][pc + 1]
            elif op in (6, 7, 8):  # 6 - arc   7 - filled arc   8 - hollow arc
                center = (x, y)

                start_angle = marker[mtype][pc + 1] * pi / 180
                end_angle = marker[mtype][pc + 2] * pi / 180

                self.write('c.beginPath();\n')
                self.draw_arc(x, y, r, start_angle, end_angle)
                self.write('c.closePath();\n')
                if not op == 6:
                    if op == 8:
                        self.write('c.fillStyle = "#FFFFFF";\n')
                    self.write('c.fill();\n')
                    self.write('c.fillStyle = "#000000";\n')
                pc += 2
            pc += 1
            if op == 0:
                break

    def set_fill_style_index(self, index):
        self.gkss.styli = index

    def set_fill_int_style(self, istyle):
        self.gkss.ints = istyle

    def set_fill_color_index(self, idx):
        self.gkss.facoli = idx

    def set_pline_linewidth(self, lwidth):
        self.gkss.lwidth = lwidth

    def set_pline_color_index(self, idx):
        self.gkss.plcoli = idx

    def set_pline_linetype(self, ltype):
        self.gkss.ltype = ltype

    def set_pmark_type(self, type):
        self.gkss.mtype = type

    def set_pmark_size(self, f):
        self.gkss.mszsc = f

    def set_pmark_color_index(self, idx):
        self.gkss.pmcoli = idx

    def draw_arc(self, x, y, r, start_angle, end_angle):
        self.write('c.beginPath();\n')
        self.write('c.arc({0},{1},{2},{3}, {4}, true);\n'.format(x, y, r, start_angle, end_angle))
        self.write('c.stroke();\n')

    def polymarker(self, n, px, py):
        if self.p.dashes != []:
            self.p.dashes = []
            self.write('set_dashes(c, []);\n')
        mk_type = self.gkss.mtype if self.gkss.asf[3] else self.gkss.mindex
        mk_size = self.gkss.mszsc if self.gkss.asf[4] else 1
        mk_color = self.gkss.pmcoli if self.gkss.asf[5] else 1
        color = self.p.rgb[mk_color]
        color.append(self.transparency)

        if self.p.fillStyle != color:
            self.p.fillStyle = color
            self.write('c.fillStyle="rgba({0},{1},{2},{3})";\n'.format(*color))
        if self.p.strokeStyle != color:
            self.p.strokeStyle = color
            self.write('c.strokeStyle="rgba({0},{1},{2},{3})";\n'.format(*color))

        ln_width = max(1, round(self.p.height / 500.0)) if self.gkss.version > 4 else 1
        if self.p.lineWidth != ln_width:
            self.p.lineWidth = ln_width
            self.write('c.lineWidth = {0};\n'.format(ln_width))

        for i in range(len(px)):
            (x, y) = self.WC_to_NDC(px[i], py[i], self.gkss.cntnr)
            (x, y) = self.seg_xform(x, y)
            self.draw_marker(x, y, mk_type, mk_size, 1)

    def seg_xform_rel(self, x, y):
        xx = x * self.gkss.mat[0][0] + y * self.gkss.mat[0][1]
        y = x * self.gkss.mat[1][0] + y * self.gkss.mat[1][1]
        return (xx, y)

    def seg_xform(self, x, y):
        xx = x * self.gkss.mat[0][0] + y * self.gkss.mat[0][1] + self.gkss.mat[2][0]
        y = x * self.gkss.mat[1][0] + y * self.gkss.mat[1][1] + self.gkss.mat[2][1]
        return (xx, y)

    def set_xform(self):
        self.p.a = (self.p.width - 1) / (self.p.window[1] - self.p.window[0])
        self.p.b = -self.p.window[0] * self.p.a
        self.p.c = (self.p.height - 1) / (self.p.window[2] - self.p.window[3])
        self.p.d = self.p.height - 1 - self.p.window[2] * self.p.c

    def set_norm_xform(self, tnr, wn, vp):
        a[tnr] = (vp[1] - vp[0]) / (wn[1] - wn[0])
        b[tnr] = vp[0] - wn[0] * a[tnr]
        c[tnr] = (vp[3] - vp[2]) / (wn[3] - wn[2])
        d[tnr] = vp[2] - wn[2] * c[tnr]

        (xp1, yp1) = self.NDC_to_DC(vp[0], vp[3])
        (xp2, yp2) = self.NDC_to_DC(vp[1], vp[2])

        self.p.rect[tnr] = ((xp1, yp1), (xp2, yp2))

    def set_text_fontprec(self, font, prec):
        self.gkss.txfont = font
        self.gkss.txprec = prec

    def set_text_expfac(self, exp_fac):
        self.gkss.chxp = exp_fac

    def set_text_spacing(self, sp):
        self.gkss.chsp = sp

    def set_text_color_index(self, index):
        self.gkss.txcoli = index

    def set_text_height(self, h):
        self.gkss.chh = h

    def set_text_upvec(self, x, y):
        self.gkss.chup[0] = x
        self.gkss.chup[1] = y

    def set_text_path(self, path):
        self.gkss.txp = path

    def set_text_align(self, i, j):
        self.gkss.txal[0] = i
        self.gkss.txal[1] = j

    def set_text_slant(self, slant):
        self.gkss.txslant = slant

    def set_clipping(self, clip):
        self.gkss.clip = clip
        self.set_clip_rect(self.gkss.cntnr)

    def set_clip_rect(self, tnr):
        if self.gkss.clip == GKS_K_CLIP and self.p.clip_rect == self.p.rect[tnr]:
            return

        if self.p.clip_rect:
            self.write('c.restore();\n')

        if self.gkss.clip == GKS_K_CLIP:
            self.p.clip_rect = self.p.rect[tnr]
            self.write('c.save();\n')
            self.write('c.beginPath();\n')
            self.write('c.rect({}, {}, {}, {});\n'.format(self.p.clip_rect[0][0], self.p.clip_rect[0][1],
                                                          self.p.clip_rect[1][0], self.p.clip_rect[1][1]))
            self.write('c.clip();\n')
        else:
            self.p.clip_rect = ()

    def set_font(self, font):

        font = abs(font)
        if font >= 101 and font <= 129:
            font -= 100
        elif font >= 1 and font <= 32:
            font = font_map[font - 1]
        else:
            font = 9

        (ux, uy) = self.WC_to_NDC_rel(self.gkss.chup[0], self.gkss.chup[1], self.gkss.cntnr)
        (ux, uy) = self.seg_xform_rel(ux, uy)
        self.p.alpha = -atan2(ux, uy)
        if self.p.alpha < 0:
            self.p.alpha += 2 * pi

        scale = sqrt(self.gkss.chup[0] * self.gkss.chup[0] + self.gkss.chup[1] * self.gkss.chup[1])
        ux = self.gkss.chup[0] / scale * self.gkss.chh
        uy = self.gkss.chup[1] / scale * self.gkss.chh
        (ux, uy) = self.WC_to_NDC_rel(ux, uy, self.gkss.cntnr)

        width = 0
        height = sqrt(ux * ux + uy * uy)
        (width, height) = self.seg_xform_rel(width, height)

        height = sqrt(width * width + height * height)
        capheight = round(height * (abs(self.p.c) + 1))
        self.p.capheight = round(capheight)

        fontNum = font - 1
        size = round(self.p.capheight / capheights[fontNum])

        if font > 13:
            font += 3
        self.p.family = (font - 1) / 4
        bold = 0 if (font % 4 == 1 or font % 4 == 2) else 1
        italic = (font % 4 == 2 or font % 4 == 0)

        font_str = ''

        if bold:
            font_str += 'bold '
        if italic:
            font_str += 'italic '

        font_str += str(size) + 'px '
        font_str += fonts[self.p.family]

        if self.p.font != font_str:
            self.p.font = font_str
            self.write('c.font = "{0}";\n'.format(font_str))

    def set_color_rep(self, color, red, green, blue):
        if color >= 0 and color < MAX_COLOR:
            self.p.rgb[color] = [int(255 * red), int(255 * green), int(255 * blue), 255]

    def set_transparency(self, val):
        self.write('c.globalAlpha = {0};'.format(val))
        self.transparency = val

    def init_colors(self):
        for color in range(MAX_COLOR):
            (red, green, blue) = gks.inq_rgb(color)
            self.set_color_rep(color, red, green, blue)

    def init_norm_xform(self):
        for tnr in range(MAX_TNR):
            self.set_norm_xform(tnr, self.gkss.window[tnr], self.gkss.viewport[tnr])

    def WC_to_NDC(self, xw, yw, tnr):
        xn = a[tnr] * xw + b[tnr]
        yn = c[tnr] * yw + d[tnr]
        return (xn, yn)

    def WC_to_NDC_rel(self, xw, yw, tnr):
        xn = a[tnr] * xw
        yn = c[tnr] * yw
        return (xn, yn)

    def NDC_to_DC(self, xn, yn):
        xd = self.p.a * xn + self.p.b
        yd = self.p.c * yn + self.p.d
        return (xd, yd)

    def DC_to_NDC(self, xd, yd):
        xn = (xd - self.p.b) / self.p.a
        yn = (yd - self.p.d) / self.p.c
        return (xn, yn)

    def select_xform(self, tnr):
        self.gkss.cntnr = tnr
        self.set_clip_rect(tnr)

    def write(self, s):
        self.file.write(self.indentation * '  ' + s)



#!/usr/bin/env python
# -*- animation -*-
"""
Play an audio file and display spectrogram in realtime
"""

import os, wave, pyaudio
import numpy as np
import gr
import gr3

FS=44100       # Sampling frequency
SAMPLES = 2048

wf = wave.open(os.path.join(os.path.dirname(os.path.realpath(__file__)),
                            'Monty_Python.wav'), 'rb')
pa = pyaudio.PyAudio()
stream = pa.open(format=pa.get_format_from_width(wf.getsampwidth()),
                 channels=wf.getnchannels(), rate=wf.getframerate(), output=True)

spectrum = np.zeros((256, 64), dtype=float)
t = -63
dt = float(SAMPLES) / FS
df = FS / float(SAMPLES) / 2 / 2

data = wf.readframes(SAMPLES)
while data != '' and len(data) == SAMPLES * wf.getsampwidth():
    stream.write(data)
    amplitudes = np.fromstring(data, dtype=np.short)
    power = abs(np.fft.fft(amplitudes / 32768.0))[:SAMPLES/2]

    gr.clearws()
    spectrum[:, 63] = power[:256]
    spectrum = np.roll(spectrum, 1)
    gr.setcolormap(-113)
    gr.setviewport(0.05, 0.95, 0.1, 1)
    gr.setwindow(t * dt, (t + 63) * dt, 0, df)
    gr.setscale(gr.OPTION_FLIP_X)
    gr.setspace(0, 256, 30, 80)
    gr3.surface((t + np.arange(64)) * dt, np.linspace(0, df, 256), spectrum, 4)
    gr.setscale(0)
    gr.axes3d(0.2, 0.2, 0, (t + 63) * dt, 0, 0, 5, 5, 0, -0.01)
    gr.titles3d('t [s]', 'f [kHz]', '')
    gr.updatews()

    data = wf.readframes(SAMPLES)
    t += 1



#!/usr/bin/env python
# -*- no-plot -*-
"""
Calculate Mandelbrot set using OpenCL (GR3 version)
"""

import pyopencl as cl
from timeit import default_timer as timer
from os import getenv, environ

import numpy as np
import gr
import gr3

def calc_fractal(q, min_x, max_x, min_y, max_y, width, height, iters):
    ctx = cl.create_some_context()
    queue = cl.CommandQueue(ctx)

    prg = cl.Program(ctx, """
    #pragma OPENCL EXTENSION cl_khr_byte_addressable_store : enable
    __kernel void mandelbrot(__global double2 *q, __global ushort *output,
                             double const min_x, double const max_x,
                             double const min_y, double const max_y,
                             ushort const width, ushort const height,
                             ushort const iters)
    {
        int ci = 0, inc = 1;
        int gid = get_global_id(0);
        double nreal, real = 0;
        double imag = 0;

        q[gid].x = min_x + (gid % width) * (max_x - min_x) / width;
        q[gid].y = min_y + (gid / width) * (max_y - min_y) / height;

        output[gid] = iters;

        for (int curiter = 0; curiter < iters; curiter++) {
            nreal = real * real - imag * imag + q[gid].x;
            imag = 2 * real * imag + q[gid].y;
            real = nreal;

            if (real * real + imag * imag >= 4) {
                 output[gid] = ci;
                 return;
            }
            ci += inc;
            if (ci == 0 || ci == 255)
                inc = -inc;
        }
    }
    """).build()

    output = np.empty(q.shape, dtype=np.uint16)

    mf = cl.mem_flags
    q_opencl = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=q)
    output_opencl = cl.Buffer(ctx, mf.WRITE_ONLY, output.nbytes)

    prg.mandelbrot(queue, output.shape, None, q_opencl, output_opencl,
                   np.double(min_x), np.double(max_x),
                   np.double(min_y), np.double(max_y),
                   np.uint16(width), np.uint16(height), np.uint16(iters))

    cl.enqueue_copy(queue, output, output_opencl).wait()

    return output

def create_fractal(min_x, max_x, min_y, max_y, width, height, iters):
    q = np.zeros(width * height).astype(np.complex128)

    output = calc_fractal(q, min_x, max_x, min_y, max_y, width, height, iters)

    return output


if getenv('PYOPENCL_CTX') == None:
    environ['PYOPENCL_CTX'] = '0'

x = -0.9223327810370947027656057193752719757635
y = 0.3102598350874576432708737495917724836010

f = 0.5
for i in range(200):
    start = timer()
    pixels = create_fractal(x-f, x+f, y-f, y+f, 500, 500, 400)
    dt = timer() - start

    print("Mandelbrot created in %f s" % dt)

    gr.clearws()
    gr.setviewport(0, 1, 0, 1)
    gr.setcolormap(113)
    z = np.resize(pixels, (500, 500))
    gr.setwindow(0, 500, 0, 500)
    gr.setspace(0, 600, 30, 80)
    gr3.surface(range(500), range(500), z, 3)
    gr.updatews()

    f *= 0.9



# -*- coding: utf-8 -*-
"""
This module includes information about elements:
- atom_name_list maps element number to name
- atom_color_list maps element number to rgb color tuple
- atomic_number_dict maps element symbol to element number (not a 1:1, but a n:1 mapping, see Ununbium (UUB)/Copernicum(CN))
- atom_radius_list maps element number to covalent radius in mÅ
"""

atom_name_list = [None,"Hydrogen","Helium","Lithium","Beryllium","Boron","Carbon","Nitrogen","Oxygen","Fluorine","Neon","Sodium","Magnesium","Aluminium","Silicon","Phosphorus","Sulfur","Chlorine","Argon","Potassium","Calcium","Scandium","Titanium","Vanadium","Chromium","Manganese","Iron","Cobalt","Nickel","Copper","Zinc","Gallium","Germanium","Arsenic","Selenium","Bromine","Krypton","Rubidium","Strontium","Yttrium","Zirconium","Niobium","Molybdenum","Technetium","Ruthenium","Rhodium","Palladium","Silver","Cadmium","Indium","Tin","Antimony","Tellurium","Iodine","Xenon","Caesium","Barium","Lanthanum","Cerium","Praseodymium","Neodymium","Promethium","Samarium","Europium","Gadolinium","Terbium","Dysprosium","Holmium","Erbium","Thulium","Ytterbium","Lutetium","Hafnium","Tantalum","Tungsten","Rhenium","Osmium","Iridium","Platinum","Gold","Mercury","Thallium","Lead","Bismuth","Polonium","Astatine","Radon","Francium","Radium","Actinium","Thorium","Protactinium","Uranium","Neptunium","Plutonium","Americium","Curium","Berkelium","Californium","Einsteinium","Fermium","Mendelevium","Nobelium","Lawrencium","Rutherfordium","Dubnium","Seaborgium","Bohrium","Hassium","Meitnerium","Darmstadtium","Roentgenium","Copernicium","Ununtrium","Ununquadium","Ununpentium","Ununhexium","Ununseptium","Ununoctium"]
atom_color_list = [None]*119
atomic_number_dict = {}
atomic_number_dict["H"] = 1
atom_color_list[ 1 ] = (255, 255, 255)
atomic_number_dict["HE"] = 2
atom_color_list[ 2 ] = (217, 255, 255)
atomic_number_dict["LI"] = 3
atom_color_list[ 3 ] = (204, 128, 255)
atomic_number_dict["BE"] = 4
atom_color_list[ 4 ] = (194, 255, 0)
atomic_number_dict["B"] = 5
atom_color_list[ 5 ] = (255, 181, 181)
atomic_number_dict["C"] = 6
atom_color_list[ 6 ] = (144, 144, 144)
atomic_number_dict["N"] = 7
atom_color_list[ 7 ] = (48, 80, 248)
atomic_number_dict["O"] = 8
atom_color_list[ 8 ] = (255, 13, 13)
atomic_number_dict["F"] = 9
atom_color_list[ 9 ] = (144, 224, 80)
atomic_number_dict["NE"] = 10
atom_color_list[ 10 ] = (179, 227, 245)
atomic_number_dict["NA"] = 11
atom_color_list[ 11 ] = (171, 92, 242)
atomic_number_dict["MG"] = 12
atom_color_list[ 12 ] = (138, 255, 0)
atomic_number_dict["AL"] = 13
atom_color_list[ 13 ] = (191, 166, 166)
atomic_number_dict["SI"] = 14
atom_color_list[ 14 ] = (240, 200, 160)
atomic_number_dict["P"] = 15
atom_color_list[ 15 ] = (255, 128, 0)
atomic_number_dict["S"] = 16
atom_color_list[ 16 ] = (255, 255, 48)
atomic_number_dict["CL"] = 17
atom_color_list[ 17 ] = (31, 240, 31)
atomic_number_dict["AR"] = 18
atom_color_list[ 18 ] = (128, 209, 227)
atomic_number_dict["K"] = 19
atom_color_list[ 19 ] = (143, 64, 212)
atomic_number_dict["CA"] = 20
atom_color_list[ 20 ] = (61, 225, 0)
atomic_number_dict["SC"] = 21
atom_color_list[ 21 ] = (230, 230, 230)
atomic_number_dict["TI"] = 22
atom_color_list[ 22 ] = (191, 194, 199)
atomic_number_dict["V"] = 23
atom_color_list[ 23 ] = (166, 166, 171)
atomic_number_dict["CR"] = 24
atom_color_list[ 24 ] = (138, 153, 199)
atomic_number_dict["MN"] = 25
atom_color_list[ 25 ] = (156, 122, 199)
atomic_number_dict["FE"] = 26
atom_color_list[ 26 ] = (224, 102, 51)
atomic_number_dict["CO"] = 27
atom_color_list[ 27 ] = (240, 144, 160)
atomic_number_dict["NI"] = 28
atom_color_list[ 28 ] = (80, 208, 80)
atomic_number_dict["CU"] = 29
atom_color_list[ 29 ] = (200, 128, 51)
atomic_number_dict["ZN"] = 30
atom_color_list[ 30 ] = (125, 128, 176)
atomic_number_dict["GA"] = 31
atom_color_list[ 31 ] = (194, 143, 143)
atomic_number_dict["GE"] = 32
atom_color_list[ 32 ] = (102, 143, 143)
atomic_number_dict["AS"] = 33
atom_color_list[ 33 ] = (189, 128, 227)
atomic_number_dict["SE"] = 34
atom_color_list[ 34 ] = (225, 161, 0)
atomic_number_dict["BR"] = 35
atom_color_list[ 35 ] = (166, 41, 41)
atomic_number_dict["KR"] = 36
atom_color_list[ 36 ] = (92, 184, 209)
atomic_number_dict["RB"] = 37
atom_color_list[ 37 ] = (112, 46, 176)
atomic_number_dict["SR"] = 38
atom_color_list[ 38 ] = (0, 255, 0)
atomic_number_dict["Y"] = 39
atom_color_list[ 39 ] = (148, 255, 255)
atomic_number_dict["ZR"] = 40
atom_color_list[ 40 ] = (148, 224, 224)
atomic_number_dict["NB"] = 41
atom_color_list[ 41 ] = (115, 194, 201)
atomic_number_dict["MO"] = 42
atom_color_list[ 42 ] = (84, 181, 181)
atomic_number_dict["TC"] = 43
atom_color_list[ 43 ] = (59, 158, 158)
atomic_number_dict["RU"] = 44
atom_color_list[ 44 ] = (36, 143, 143)
atomic_number_dict["RH"] = 45
atom_color_list[ 45 ] = (10, 125, 140)
atomic_number_dict["PD"] = 46
atom_color_list[ 46 ] = (0, 105, 133)
atomic_number_dict["AG"] = 47
atom_color_list[ 47 ] = (192, 192, 192)
atomic_number_dict["CD"] = 48
atom_color_list[ 48 ] = (255, 217, 143)
atomic_number_dict["IN"] = 49
atom_color_list[ 49 ] = (166, 117, 115)
atomic_number_dict["SN"] = 50
atom_color_list[ 50 ] = (102, 128, 128)
atomic_number_dict["SB"] = 51
atom_color_list[ 51 ] = (158, 99, 181)
atomic_number_dict["TE"] = 52
atom_color_list[ 52 ] = (212, 122, 0)
atomic_number_dict["I"] = 53
atom_color_list[ 53 ] = (148, 0, 148)
atomic_number_dict["XE"] = 54
atom_color_list[ 54 ] = (66, 158, 176)
atomic_number_dict["CS"] = 55
atom_color_list[ 55 ] = (87, 23, 143)
atomic_number_dict["BA"] = 56
atom_color_list[ 56 ] = (0, 201, 0)
atomic_number_dict["LA"] = 57
atom_color_list[ 57 ] = (112, 212, 255)
atomic_number_dict["CE"] = 58
atom_color_list[ 58 ] = (255, 255, 199)
atomic_number_dict["PR"] = 59
atom_color_list[ 59 ] = (217, 225, 199)
atomic_number_dict["ND"] = 60
atom_color_list[ 60 ] = (199, 225, 199)
atomic_number_dict["PM"] = 61
atom_color_list[ 61 ] = (163, 225, 199)
atomic_number_dict["SM"] = 62
atom_color_list[ 62 ] = (143, 225, 199)
atomic_number_dict["EU"] = 63
atom_color_list[ 63 ] = (97, 225, 199)
atomic_number_dict["GD"] = 64
atom_color_list[ 64 ] = (69, 225, 199)
atomic_number_dict["TB"] = 65
atom_color_list[ 65 ] = (48, 225, 199)
atomic_number_dict["DY"] = 66
atom_color_list[ 66 ] = (31, 225, 199)
atomic_number_dict["HO"] = 67
atom_color_list[ 67 ] = (0, 225, 156)
atomic_number_dict["ER"] = 68
atom_color_list[ 68 ] = (0, 230, 117)
atomic_number_dict["TM"] = 69
atom_color_list[ 69 ] = (0, 212, 82)
atomic_number_dict["YB"] = 70
atom_color_list[ 70 ] = (0, 191, 56)
atomic_number_dict["LU"] = 71
atom_color_list[ 71 ] = (0, 171, 36)
atomic_number_dict["HF"] = 72
atom_color_list[ 72 ] = (77, 194, 255)
atomic_number_dict["TA"] = 73
atom_color_list[ 73 ] = (77, 166, 255)
atomic_number_dict["W"] = 74
atom_color_list[ 74 ] = (33, 148, 214)
atomic_number_dict["RE"] = 75
atom_color_list[ 75 ] = (38, 125, 171)
atomic_number_dict["OS"] = 76
atom_color_list[ 76 ] = (38, 102, 150)
atomic_number_dict["IR"] = 77
atom_color_list[ 77 ] = (23, 84, 135)
atomic_number_dict["PT"] = 78
atom_color_list[ 78 ] = (208, 208, 224)
atomic_number_dict["AU"] = 79
atom_color_list[ 79 ] = (255, 209, 35)
atomic_number_dict["HG"] = 80
atom_color_list[ 80 ] = (184, 184, 208)
atomic_number_dict["TL"] = 81
atom_color_list[ 81 ] = (166, 84, 77)
atomic_number_dict["PB"] = 82
atom_color_list[ 82 ] = (87, 89, 97)
atomic_number_dict["BI"] = 83
atom_color_list[ 83 ] = (158, 79, 181)
atomic_number_dict["PO"] = 84
atom_color_list[ 84 ] = (171, 92, 0)
atomic_number_dict["AT"] = 85
atom_color_list[ 85 ] = (117, 79, 69)
atomic_number_dict["RN"] = 86
atom_color_list[ 86 ] = (66, 130, 150)
atomic_number_dict["FR"] = 87
atom_color_list[ 87 ] = (66, 0, 102)
atomic_number_dict["RA"] = 88
atom_color_list[ 88 ] = (0, 125, 0)
atomic_number_dict["AC"] = 89
atom_color_list[ 89 ] = (112, 171, 250)
atomic_number_dict["TH"] = 90
atom_color_list[ 90 ] = (0, 186, 255)
atomic_number_dict["PA"] = 91
atom_color_list[ 91 ] = (0, 161, 255)
atomic_number_dict["U"] = 92
atom_color_list[ 92 ] = (0, 143, 255)
atomic_number_dict["NP"] = 93
atom_color_list[ 93 ] = (0, 128, 255)
atomic_number_dict["PU"] = 94
atom_color_list[ 94 ] = (0, 107, 255)
atomic_number_dict["AM"] = 95
atom_color_list[ 95 ] = (84, 92, 242)
atomic_number_dict["CM"] = 96
atom_color_list[ 96 ] = (120, 92, 227)
atomic_number_dict["BK"] = 97
atom_color_list[ 97 ] = (138, 79, 227)
atomic_number_dict["CF"] = 98
atom_color_list[ 98 ] = (161, 54, 212)
atomic_number_dict["ES"] = 99
atom_color_list[ 99 ] = (179, 31, 212)
atomic_number_dict["FM"] = 100
atom_color_list[ 100 ] = (179, 31, 186)
atomic_number_dict["MD"] = 101
atom_color_list[ 101 ] = (179, 13, 166)
atomic_number_dict["NO"] = 102
atom_color_list[ 102 ] = (189, 13, 135)
atomic_number_dict["LR"] = 103
atom_color_list[ 103 ] = (199, 0, 102)
atomic_number_dict["RF"] = 104
atom_color_list[ 104 ] = (204, 0, 89)
atomic_number_dict["DB"] = 105
atom_color_list[ 105 ] = (209, 0, 79)
atomic_number_dict["SG"] = 106
atom_color_list[ 106 ] = (217, 0, 69)
atomic_number_dict["BH"] = 107
atom_color_list[ 107 ] = (224, 0, 56)
atomic_number_dict["HS"] = 108
atom_color_list[ 108 ] = (230, 0, 46)
atomic_number_dict["MT"] = 109
atom_color_list[ 109 ] = (235, 0, 38)
atomic_number_dict["DS"] = 110
atom_color_list[ 110 ] = (255, 0, 255)
atomic_number_dict["RG"] = 111
atom_color_list[ 111 ] = (255, 0, 255)
atomic_number_dict["CN"] = 112
atomic_number_dict["UUB"] = 112
atom_color_list[ 112 ] = (255, 0, 255)
atomic_number_dict["UUT"] = 113
atom_color_list[ 113 ] = (255, 0, 255)
atomic_number_dict["UUQ"] = 114
atom_color_list[ 114 ] = (255, 0, 255)
atomic_number_dict["UUP"] = 115
atom_color_list[ 115 ] = (255, 0, 255)
atomic_number_dict["UUH"] = 116
atom_color_list[ 116 ] = (255, 0, 255)
atomic_number_dict["UUS"] = 117
atom_color_list[ 117 ] = (255, 0, 255)
atomic_number_dict["UUO"] = 118
atom_color_list[ 118 ] = (255, 0, 255)

# Source: Jmol constants, units: mÅ
atom_radius_list = [ None,
230,  #   1  H
930,  #   2  He
680,  #   3  Li
350,  #   4  Be
830,  #   5  B
680,  #   6  C
680,  #   7  N
680,  #   8  O
640,  #   9  F
1120, #  10  Ne
970,  #  11  Na
1100, #  12  Mg
1350, #  13  Al
1200, #  14  Si
750,  #  15  P
1020, #  16  S
990,  #  17  Cl
1570, #  18  Ar
1330, #  19  K
990,  #  20  Ca
1440, #  21  Sc
1470, #  22  Ti
1330, #  23  V
1350, #  24  Cr
1350, #  25  Mn
1340, #  26  Fe
1330, #  27  Co
1500, #  28  Ni
1520, #  29  Cu
1450, #  30  Zn
1220, #  31  Ga
1170, #  32  Ge
1210, #  33  As
1220, #  34  Se
1210, #  35  Br
1910, #  36  Kr
1470, #  37  Rb
1120, #  38  Sr
1780, #  39  Y
1560, #  40  Zr
1480, #  41  Nb
1470, #  42  Mo
1350, #  43  Tc
1400, #  44  Ru
1450, #  45  Rh
1500, #  46  Pd
1590, #  47  Ag
1690, #  48  Cd
1630, #  49  In
1460, #  50  Sn
1460, #  51  Sb
1470, #  52  Te
1400, #  53  I
1980, #  54  Xe
1670, #  55  Cs
1340, #  56  Ba
1870, #  57  La
1830, #  58  Ce
1820, #  59  Pr
1810, #  60  Nd
1800, #  61  Pm
1800, #  62  Sm
1990, #  63  Eu
1790, #  64  Gd
1760, #  65  Tb
1750, #  66  Dy
1740, #  67  Ho
1730, #  68  Er
1720, #  69  Tm
1940, #  70  Yb
1720, #  71  Lu
1570, #  72  Hf
1430, #  73  Ta
1370, #  74  W
1350, #  75  Re
1370, #  76  Os
1320, #  77  Ir
1500, #  78  Pt
1500, #  79  Au
1700, #  80  Hg
1550, #  81  Tl
1540, #  82  Pb
1540, #  83  Bi
1680, #  84  Po
1700, #  85  At
2400, #  86  Rn
2000, #  87  Fr
1900, #  88  Ra
1880, #  89  Ac
1790, #  90  Th
1610, #  91  Pa
1580, #  92  U
1550, #  93  Np
1530, #  94  Pu
1510, #  95  Am
1500, #  96  Cm
1500, #  97  Bk
1500, #  98  Cf
1500, #  99  Es
1500, # 100  Fm
1500, # 101  Md
1500, # 102  No
1500, # 103  Lr
1600, # 104  Rf
1600, # 105  Db
1600, # 106  Sg
1600, # 107  Bh
1600, # 108  Hs
1600, # 109  Mt

1600, # 110 DS
1600, # 111 RG
1600, # 112 CN
1600, # 113 UUT
1600, # 114 UUQ
1600, # 115 UUP
1600, # 116 UUH
1600, # 117 UUS
1600, # 118 UUO

]



from ctypes import *
from constants import *

line_callback = None
fill_callback = None

fill_prototype = CFUNCTYPE(
    c_void_p, c_int, POINTER(c_double), POINTER(c_double), c_int)
line_prototype = CFUNCTYPE(
    c_void_p, c_int, POINTER(c_double), POINTER(c_double), c_int, c_int)

_gks = CDLL('/usr/local/gr/lib/libGKS.so')


def set_line_callback(lrc):
    global line_callback
    line_callback = line_prototype(lrc)


def set_fill_callback(frc):
    global fill_callback
    fill_callback = fill_prototype(frc)


class state_list(Structure):
    _fields_ = [
        ('lindex', c_int),
        ('ltype', c_int),
        ('lwidth', c_double),
        ('plcoli', c_int),
        ('mindex', c_int),
        ('mtype', c_int),
        ('mszsc', c_double),
        ('pmcoli', c_int),
        ('tindex', c_int),
        ('txfont', c_int),
        ('txprec', c_int),
        ('chxp', c_double),
        ('chsp', c_double),
        ('txcoli', c_int),
        ('chh', c_double),
        ('chup', c_double * 2),
        ('txp', c_int),
        ('txal', c_int * 2),
        ('findex', c_int),
        ('ints', c_int),
        ('styli', c_int),
        ('facoli', c_int),
        ('window', (c_double * 4) * MAX_TNR),
        ('viewport', (c_double * 4) * MAX_TNR),
        ('cntnr', c_int),
        ('clip', c_int),
        ('opsg', c_int),
        ('mat', (c_double * 2) * 3),
        ('asf', c_int * 13),
        ('wiss', c_int),
        ('version', c_int),
        ('fontfile', c_int),
        ('txslant', c_double),
        ('shoff', c_double * 2),
        ('blur', c_double),
        ('alpha', c_double),
        ('a', c_double * MAX_TNR),
        ('b', c_double * MAX_TNR),
        ('c', c_double * MAX_TNR),
        ('d', c_double * MAX_TNR)
    ]

    def sizeof(self):
        return sizeof(self)

    def copy_data(self, bytes):
        memmove(addressof(self), bytes, sizeof(self))

_gks.gks_set_dev_xform.argtypes = (POINTER(state_list), POINTER(c_double),
                                   POINTER(c_double))
_gks.gks_inq_pattern_array.argtypes = (c_int, c_int * 33)
_gks.gks_inq_rgb.argtypes = (
    c_int, POINTER(c_double), POINTER(c_double), POINTER(c_double))
_gks.gks_get_dash_list.argtypes = (c_int, c_double, c_int * 10)
_gks.gks_init_core.argtypes = (POINTER(state_list),)
_gks.gks_open_font.argtypes = []
_gks.gks_emul_text.argtypes = (
    c_double, c_double, c_int, c_char_p, line_prototype, fill_prototype)
_gks.gks_set_norm_xform.argtypes = (
    c_int, POINTER(c_double), POINTER(c_double))


def set_dev_xform(s_list, window, viewport):

    c_window = (c_double*len(window))()
    for i in range(len(window)):
        c_window[i] = c_double(window[i])

    c_viewport = (c_double*len(viewport))()
    for i in range(len(viewport)):
        c_viewport[i] = c_double(viewport[i])

    _gks.gks_set_dev_xform(s_list,
                           cast(c_window, POINTER(c_double)),
                           cast(c_viewport, POINTER(c_double)))


def inq_pattern_array(index):
    global _gks
    pattern = (c_int * 33)()
    _gks.gks_inq_pattern_array(c_int(index), pattern)
    n = pattern[0]
    return [i for i in pattern[1:n+1]]


def inq_rgb(color_index):
    global _gks
    color = [c_double(0) for i in range(3)]
    _gks.gks_inq_rgb(c_int(color_index), *map(byref, color))
    return [c.value for c in color]


def get_dash_list(ltype, scale):
    global _gks

    lt = c_int * 10
    l = lt()
    _gks.gks_get_dash_list(c_int(ltype), c_double(scale), l)

    l = [i for i in l[1:] if i != 0]
    return l


def init_core(list):
    global _gks

    _gks.gks_init_core(byref(list))


def open_font():
    global _gks

    fd = _gks.gks_open_font()
    return fd


def emul_text(x_pos, y_pos, text):
    _gks.gks_emul_text(c_double(x_pos), c_double(y_pos),
                       c_int(len(text)), c_char_p(text),
                       line_callback, fill_callback)


def set_norm_xform(tnr, window, viewport):
    c_window = (c_double*len(window))()
    for i in range(len(window)):
        c_window[i] = c_double(window[i])

    c_viewport = (c_double*len(viewport))()
    for i in range(len(viewport)):
        c_viewport[i] = c_double(viewport[i])

    _gks.gks_set_norm_xform(c_int(tnr),
                            cast(c_window, POINTER(c_double)),
                            cast(c_viewport, POINTER(c_double)))



#!/usr/bin/env python
# -*- no-plot -*-
"""
Calculate Mandelbrot set using Numba
"""

from numba import autojit
from timeit import default_timer as timer

from numpy import zeros, uint8
import gr

@autojit
def mandel(x, y, iters):
    c = complex(x, y)
    z = 0.0j
    ci = 0
    inc = 1

    for i in range(iters):
        z = z * z + c
        if (z.real * z.real + z.imag * z.imag) >= 4:
            return ci
        ci += inc
        if ci == 0 or ci == 255:
            inc = -inc

    return 255

@autojit
def create_fractal(min_x, max_x, min_y, max_y, image, iters):
    height = image.shape[0]
    width = image.shape[1]

    pixel_size_x = (max_x - min_x) / width
    pixel_size_y = (max_y - min_y) / height
    for x in range(width):
        real = min_x + x * pixel_size_x
        for y in range(height):
            imag = min_y + y * pixel_size_y
            color = mandel(real, imag, iters)
            image[y, x] = color

    return image

x = -0.9223327810370947027656057193752719757635
y = 0.3102598350874576432708737495917724836010

f = 0.5
for i in range(200):
    start = timer()
    image = zeros((500, 500), dtype=uint8)
    pixels = create_fractal(x-f, x+f, y-f, y+f, image, 400)
    dt = timer() - start

    print("Mandelbrot created in %f s" % dt)
    ca = 1000.0 + pixels.ravel()

    gr.clearws()
    gr.setviewport(0, 1, 0, 1)
    gr.setcolormap(13)
    gr.cellarray(0, 1, 0, 1, 500, 500, ca)
    gr.updatews()

    f *= 0.9



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""PyQt GR module

Exported Classes:

"""
# standard library
import os
import math
import logging
# third party
getGKSConnectionId = None
def _getGKSConnectionIdPySide(widget, painter):
    return "%x!%x" % (long(shiboken.getCppPointer(widget)[0]),
                      long(shiboken.getCppPointer(painter)[0]))

def _getGKSConnectionIdPyQt4(widget, painter):
    return "%x!%x" % (sip.unwrapinstance(widget),
                      sip.unwrapinstance(painter))

def _importPySide():
    global QtCore, QtGui, shiboken, getGKSConnectionId
    from PySide import QtCore
    from PySide import QtGui
    try:
        from PySide import shiboken
    except ImportError:
        import shiboken # Anaconda
    getGKSConnectionId = _getGKSConnectionIdPySide

def _importPyQt4():
    global QtCore, QtGui, sip, getGKSConnectionId
    from PyQt4 import QtCore
    from PyQt4 import QtGui
    import sip
    getGKSConnectionId = _getGKSConnectionIdPyQt4
    QtCore.Signal = QtCore.pyqtSignal

from qtgr.backend import QT_BACKEND_ORDER, QT_PYSIDE, QT_PYQT4
_imp = {QT_PYSIDE: _importPySide,
        QT_PYQT4: _importPyQt4}
try:
    _imp[QT_BACKEND_ORDER[0]]()
except ImportError:
    _imp[QT_BACKEND_ORDER[1]]()
# local library
import gr
import qtgr.events
from gr.pygr import Plot, PlotAxes, RegionOfInterest
from qtgr.events import GUIConnector, MouseEvent, PickEvent, ROIEvent, \
    LegendEvent
from gr._version import __version__, __revision__

__author__ = "Christian Felder <c.felder@fz-juelich.de>"
__copyright__ = """Copyright 2012-2014 Forschungszentrum Juelich GmbH

This file is part of GR, a universal framework for visualization applications.
Visit http://gr-framework.org for the latest
version.

GR was developed by the Scientific IT-Systems group at the Peter Grünberg
Institute at Forschunsgzentrum Jülich. The main development has been done
by Josef Heinen who currently maintains the software.

GR is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This framework is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GR. If not, see <http://www.gnu.org/licenses/>.
 
"""

_log = logging.getLogger(__name__)


class GRWidget(QtGui.QWidget):

    def __init__(self, *args, **kwargs):
        super(GRWidget, self).__init__(*args, **kwargs)
        self._clear, self._update = False, False
        self._sizex, self._sizey = 1., 1.
        self._dwidth, self._dheight = self.width(), self.height()
        self._mwidth = self.width() * .0254 / self.logicalDpiX()
        self._mheight = self.height() * .0254 / self.logicalDpiY()
        self._keepRatio = False
        self._bgColor = QtCore.Qt.white
        os.environ["GKS_WSTYPE"] = "381" # GKS Qt Plugin
        os.environ["GKS_DOUBLE_BUF"] = "True"

    def paintEvent(self, event):
        self._painter = QtGui.QPainter()
        self._painter.begin(self)
        self._painter.fillRect(0, 0, self.width(), self.height(), self._bgColor)
        os.environ["GKSconid"] = getGKSConnectionId(self, self._painter) 
        self.draw(self._clear, self._update)
        gr.updatews()
        self._painter.end()

    def resizeEvent(self, event):
        self._dwidth, self._dheight = self.width(), self.height()
        self._mwidth = self.width() * .0254 / self.logicalDpiX()
        self._mheight = self.height() * .0254 / self.logicalDpiY()
        if self._mwidth > self._mheight:
            self._sizex = 1.
            if self.keepRatio:
                self._sizey = 1.
                self._mwidth = self._mheight
                self._dwidth = self._dheight
            else:
                self._sizey = self._mheight / self._mwidth
        else:
            if self.keepRatio:
                self._sizex = 1.
                self._mheight = self._mwidth
                self._dheight = self._dwidth
            else:
                self._sizex = self._mwidth / self._mheight
            self._sizey = 1.

    def setBackground(self, qcolor):
        self._bgColor = qcolor

    @property
    def mwidth(self):
        """Get metric width of the widget excluding any window frame."""
        return self._mwidth

    @property
    def mheight(self):
        """Get metric height of the widget excluding any window frame."""
        return self._mheight

    @property
    def dwidth(self):
        """Get device width in consideration of ratio (keepRatio)."""
        return self._dwidth

    @property
    def dheight(self):
        """Get device height in consideration of ratio (keepRatio)."""
        return self._dheight

    @property
    def sizex(self):
        """..."""
        return self._sizex

    @property
    def sizey(self):
        """..."""
        return self._sizey

    @property
    def keepRatio(self):
        return self._keepRatio

    @keepRatio.setter
    def keepRatio(self, bool):
        self._keepRatio = bool
        self.resizeEvent(None)
        self.update()

    def _draw(self, clear=False, update=True):
        self._clear, self._update = clear, update

    def draw(self, clear=False, update=True):
        # put gr commands in here
        pass

    def save(self, path):
        (p, ext) = os.path.splitext(path)
        if ext.lower()[1:] == gr.GRAPHIC_GRX:
            gr.begingraphics(path)
            self.draw()
            gr.endgraphics()
        else:
            gr.beginprint(path)
            self.draw()
            gr.endprint()
        self.repaint()

    def printDialog(self, documentName="qtgr-untitled"):
        printer = QtGui.QPrinter(QtGui.QPrinter.HighResolution)
        printer.setDocName(documentName)
        painter = QtGui.QPainter()
        dlg = QtGui.QPrintDialog(printer)
        if dlg.exec_() == QtGui.QPrintDialog.Accepted:
            painter.begin(printer)
            os.environ["GKSconid"] = "%x!%x" % (sip.unwrapinstance(self),
                                               sip.unwrapinstance(painter))

            # upscaling to paper size and
            # alignment (horizontal and vertical centering)
            xscale = printer.pageRect().width() / float(self.width());
            yscale = printer.pageRect().height() / float(self.height());
            scale = min(xscale, yscale);
            painter.translate(printer.paperRect().x() +
                              printer.pageRect().width() / 2,
                              printer.paperRect().y() +
                              printer.pageRect().height() / 2)
            painter.scale(scale, scale);
            painter.translate(-self.width() / 2, -self.height() / 2);

            self.draw(True)
            gr.updatews()
            painter.end()

    def __del__(self):
        if gr:
            gr.emergencyclosegks()
        # super destructor not available
#        super(GRWidget, self).__del__()


class InteractiveGRWidget(GRWidget):

    logXinDomain = QtCore.Signal(bool)
    logYinDomain = QtCore.Signal(bool)
    modePick = QtCore.Signal(bool)

    def __init__(self, *args, **kwargs):
        super(InteractiveGRWidget, self).__init__(*args, **kwargs)
        guiConn = GUIConnector(self)
        guiConn.connect(qtgr.events.MouseEvent.MOUSE_MOVE, self.mouseMove)
        guiConn.connect(qtgr.events.MouseEvent.MOUSE_PRESS, self.mousePress)
        guiConn.connect(qtgr.events.MouseEvent.MOUSE_RELEASE, self.mouseRelease)
        guiConn.connect(qtgr.events.WheelEvent.WHEEL_MOVE, self.wheelMove)
        guiConn.connect(qtgr.events.PickEvent.PICK_MOVE, self.pickMove)
        self.setMouseTracking(True)
        self._mouseLeft = False
        self._mouseRight = False
        self._startPoint = None
        self._curPoint = None
        self._logXinDomain = None
        self._logYinDomain = None
        self._pickMode = False
        self._pickEvent = None
        self._lstPlot = []

    def draw(self, clear=False, update=True):
        if clear:
            gr.clearws()
        gr.setwsviewport(0, self.mwidth, 0, self.mheight)
        gr.setwswindow(0, self.sizex, 0, self.sizey)

        for plot in self._lstPlot:
            plot.sizex, plot.sizey = self.sizex, self.sizey
            plot.drawGR()
            # logDomainCheck
            logXinDomain = plot.logXinDomain()
            logYinDomain = plot.logYinDomain()
            if logXinDomain != self._logXinDomain:
                self._logXinDomain = logXinDomain
                self.logXinDomain.emit(self._logXinDomain)
            if logYinDomain != self._logYinDomain:
                self._logYinDomain = logYinDomain
                self.logYinDomain.emit(self._logYinDomain)

        if self._pickEvent:
            event = self._pickEvent
            wcPoint = event.getWC(event.viewport)
            window = gr.inqwindow()
            gr.setwindow(*event.getWindow())
            gr.setmarkertype(gr.MARKERTYPE_PLUS)
            gr.polymarker([wcPoint.x], [wcPoint.y])
            gr.setwindow(*window)

    def addPlot(self, *args, **kwargs):
        for plot in args:
            if plot and plot not in self._lstPlot:
                self._lstPlot.append(plot)
        self._draw(clear=True, update=True)
        return self._lstPlot

    def plot(self, *args, **kwargs):
        plot = Plot()
        axes = PlotAxes(plot.viewport)
        axes.plot(*args, **kwargs)
        plot.addAxes(axes)
        return self.addPlot(plot)

    def paintEvent(self, event):
        super(InteractiveGRWidget, self).paintEvent(event)
        self._painter.begin(self)
        if self._mouseLeft:
            startDC = self._startPoint.getDC()
            endDC = self._curPoint.getDC()
            rect = QtCore.QRect(QtCore.QPoint(startDC.x, startDC.y),
                                QtCore.QPoint(endDC.x, endDC.y)).normalized()
            self._painter.setOpacity(.75)
            self._painter.drawRect(rect)
            self._painter.setOpacity(1.)

        self._painter.end()

    def setAutoScale(self, bool):
        for plot in self._lstPlot:
            plot.autoscale = bool

    def getPickMode(self):
        return self._pickMode

    def setPickMode(self, bool):
        self._pickMode = bool
        self.modePick.emit(self._pickMode)

    def _pick(self, p0, type):
        for plot in self._lstPlot:
            (coord, _axes, _curve) = plot.pick(p0, self.dwidth, self.dheight)
            if coord:
                dcPoint = coord.getDC()
                QtGui.QApplication.sendEvent(self, PickEvent(type,
                                                             self.dwidth,
                                                             self.dheight,
                                                             dcPoint.x,
                                                             dcPoint.y,
                                                             plot.viewport,
                                                             coord.getWindow()))

    def _select(self, p0, p1):
        self._pickEvent = None
        for plot in self._lstPlot:
            plot.select(p0, p1, self.dwidth, self.dheight)
        self._draw(True)
        self.update()

    def _pan(self, dp):
        self._pickEvent = None
        for plot in self._lstPlot:
            plot.pan(dp, self.dwidth, self.dheight)
        self._draw(True)
        self.update()

    def _zoom(self, dpercent, p0):
        self._pickEvent = None
        for plot in self._lstPlot:
            plot.zoom(dpercent, p0, self.dwidth, self.dheight)
        self._draw(True)
        self.update()

    def _roi(self, p0, type, buttons, modifiers):
        for plot in self._lstPlot:
            roi = plot.getROI(p0)
            if roi:
                if roi.regionType == RegionOfInterest.LEGEND:
                    eventObj = LegendEvent
                else:
                    eventObj = ROIEvent
                QtGui.QApplication.sendEvent(self,
                                             eventObj(type,
                                                      self.dwidth,
                                                      self.dheight,
                                                      p0.x, p0.y,
                                                      buttons, modifiers,
                                                      roi))

    def mousePress(self, event):
        if event.getButtons() & MouseEvent.LEFT_BUTTON:
            if self.getPickMode():
                self.setPickMode(False)
                self._pick(event.getNDC(), PickEvent.PICK_PRESS)
            else:
                if event.getModifiers() & MouseEvent.CONTROL_MODIFIER:
                    self.setPickMode(True)
                else:
                    self._mouseLeft = True
        elif event.getButtons() & MouseEvent.RIGHT_BUTTON:
            self._mouseRight = True
        self._curPoint = event
        self._startPoint = event

    def mouseRelease(self, event):
        if event.getButtons() & MouseEvent.LEFT_BUTTON and self._mouseLeft:
            self._mouseLeft = False
            self._curPoint = event
            p0 = self._startPoint.getNDC()
            p1 = self._curPoint.getNDC()
            if p0 != p1:
                self._select(p0, p1)
            else:
                self._roi(p0, ROIEvent.ROI_CLICKED, event.getButtons(),
                          event.getModifiers())
        elif event.getButtons() & MouseEvent.RIGHT_BUTTON:
            self._mouseRight = False
            self._roi(event.getNDC(), ROIEvent.ROI_CLICKED, event.getButtons(),
                      event.getModifiers())
        self._curPoint = event

    def mouseMove(self, event):
        if self.getPickMode():
            self._pick(event.getNDC(), PickEvent.PICK_MOVE)
        if event.getButtons() & MouseEvent.LEFT_BUTTON:
            self._curPoint = event
            super(InteractiveGRWidget, self).update()
        elif event.getButtons() & MouseEvent.RIGHT_BUTTON:
            p0 = self._curPoint.getNDC() # point before now
            p1 = event.getNDC()
            dp = p1 - p0
            self._curPoint = event
            self._pan(dp)
        self._roi(event.getNDC(), ROIEvent.ROI_OVER, event.getButtons(),
                  event.getModifiers())

    def wheelMove(self, event):
        # delta percent
        dpercent = event.getSteps() * .1
        self._zoom(dpercent, event.getNDC())

    def pickMove(self, event):
        self._pickEvent = event
        self.update()


if __name__ == "__main__":
    import sys
    from gr import pygr
    logging.basicConfig(level=logging.CRITICAL)
    for name in [__name__, pygr.base.__name__, pygr.__name__]:
        logging.getLogger(name).setLevel(logging.DEBUG)
    app = QtGui.QApplication(sys.argv)
    grw = InteractiveGRWidget()
    grw.resize(QtCore.QSize(500, 500))
    viewport = [0.1, 0.9, 0.1, 0.88]

    x = [-3.3 + t * .1 for t in range(66)]
    y = [t ** 5 - 13 * t ** 3 + 36 * t for t in x]

    n = 100
    pi2_n = 2.*math.pi / n
    x2 = [i * pi2_n for i in range(0, n + 1)]
    y2 = map(lambda xi: math.sin(xi), x2)

    plot = Plot(viewport).addAxes(PlotAxes(viewport).plot(x, y),
                                  PlotAxes(viewport).plot(x2, y2))
    plot.title, plot.subTitle = "foo", "bar"
    plot.xlabel, plot.ylabel = "x", "f(x)"

    grw.addPlot(plot)
    grw.show()
    grw.update()

    sys.exit(app.exec_())



#!/usr/bin/env python
"""
Create a contour plot of irregular distributed data
"""

from numpy.random import uniform, seed
import numpy as np
import gr

seed(0)
xd = uniform(-2, 2, 100)
yd = uniform(-2, 2, 100)
zd = xd * np.exp(-xd**2 - yd**2)

gr.setviewport(0.1, 0.95, 0.1, 0.95)
gr.setwindow(-2, 2, -2, 2)
gr.setspace(-0.5, 0.5, 0, 90)
gr.setmarkersize(1)
gr.setmarkertype(gr.MARKERTYPE_SOLID_CIRCLE)
gr.setcharheight(0.024)
gr.settextalign(2, 0)
gr.settextfontprec(3, 0)

x, y, z = gr.gridit(xd, yd, zd, 200, 200)
h = np.linspace(-0.5, 0.5, 20)
gr.surface(x, y, z, 5)
gr.contour(x, y, h, z, 0)
gr.polymarker(xd, yd)
gr.axes(0.25, 0.25, -2, -2, 2, 2, 0.01)

gr.updatews()



#!/usr/bin/env python
"""
Simple surface plot example
"""

from gr import * 
from math import *

x = [-2 + i * 0.5 for i in range(0, 29)]
y = [-7 + i * 0.5 for i in range(0, 29)]
z = list(range(0, 841))

for i in range(0, 29):
  for j in range(0, 29):
    r1 = sqrt((x[j] - 5)**2 + y[i]**2)
    r2 = sqrt((x[j] + 5)**2 + y[i]**2)
    z[i * 29 - 1 + j] = (exp(cos(r1)) + exp(cos(r2)) - 0.9) * 25

setcharheight(24.0/500)
settextalign(TEXT_HALIGN_CENTER, TEXT_VALIGN_TOP)
textext(0.5, 0.9, "Surface Example")
(tbx, tby) = inqtextext(0.5, 0.9, "Surface Example")
fillarea(tbx, tby)

setwindow(-2, 12, -7, 7)
setspace(-80, 200, 45, 70)

setcharheight(14.0/500)
axes3d(1, 0, 20, -2, -7, -80, 2, 0, 2, -0.01)
axes3d(0, 1,  0, 12, -7, -80, 0, 2, 0,  0.01)
titles3d("X-Axis", "Y-Axis", "Z-Axis")

surface(x, y, z, 3)
surface(x, y, z, 1)

updatews()



from pylab import *
A = rand(5,5)
figure(1)
imshow(A, interpolation='nearest')
grid(True)

figure(2)
imshow(A, interpolation='bilinear')
grid(True)

figure(3)
imshow(A, interpolation='bicubic')
grid(True)

show()



#!/usr/bin/env python
# -*- animation -*-
"""
Sample microphone input and display power spectrum
"""

import pyaudio
import numpy
from scipy import signal
import time
import gr

FS=44100       # Sampling frequency
SAMPLES=4096

mic = None
def get_spectrum():
    global mic
    if mic is None:
        pa = pyaudio.PyAudio()
        mic = pa.open(format=pyaudio.paInt16, channels=1, rate=FS,
                      input=True, frames_per_buffer=SAMPLES)
    amplitudes = numpy.fromstring(mic.read(SAMPLES), dtype=numpy.short)
    return abs(numpy.fft.fft(amplitudes / 32768.0))[:SAMPLES/2]

def parabolic(x, f, i):
    xe = 1/2. * (f[i-1] - f[i+1]) / (f[i-1] - 2 * f[i] + f[i+1]) + x
    ye = f[i] - 1/4. * (f[i-1] - f[i+1]) * (xe - x)
    return xe, ye

f = [FS/float(SAMPLES)*t for t in range(0, SAMPLES/2)]

gr.setviewport(0.1, 0.95, 0.1, 0.95)
gr.setwindow(50, 25000, 0, 100)
gr.setscale(1)

start = time.time()

while time.time() - start < 10:
    try:
        power = get_spectrum()
        peakind = signal.find_peaks_cwt(power, numpy.array([5]))
    except (IOError):
        continue

    gr.clearws()
    gr.setlinewidth(1)
    gr.setlinecolorind(1)
    gr.grid(1, 5, 50, 0, 1, 2)
    gr.axes(1, 5, 50, 0, 1, 2, -0.008)
    gr.setcharheight(0.020)
    gr.text(0.15, 0.965, '100Hz')
    gr.text(0.47, 0.965, '1kHz')
    gr.text(0.79, 0.965, '10kHz')
    gr.setlinecolorind(4)
    gr.polyline(f[1:], power[1:])
    for p in peakind:
        if power[p] > 10:
            gr.setlinewidth(2)
            gr.setlinecolorind(2)
            xe, ye = parabolic(f[p], power, p)
            print xe, ye
            gr.polyline([xe] * 2, [0, ye])
    gr.updatews()

